#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Samvaran Kashyap Rallabandi -  <srallaba@redhat.com>
#
# Ansible module to provision, deprovision openstack hot templates

import os
import ast
from openstack import connection

from ansible.module_utils.basic import AnsibleModule


DOCUMENTATION = '''
---
version_added: "0.1"
module: os_hot_stack
short_description: ansible openstack module to provision hot templates
description:
  - This module allows a user to provision hot templates.

options:
  stack_name:
    description:
     name of the stack to be created
    required: true
  template:
    description:
      path to topology file can be in json/yaml format
    required: true
  parameters:
    description:
      parameters passed to the hot template
  state:
    description:
      state of the hot stack
  wait:
    description:
      wait for the template provision to happen

author: Samvaran Kashyap Rallabandi
'''


def check_file_paths(module, *args):
    for file_path in args:
        if not os.path.exists(file_path):
            msg = "File not found %s not found" % (file_path)
            module.fail_json(msg=msg)
        if not os.access(file_path, os.R_OK):
            msg = "File not accesible %s not found" % (file_path)
            module.fail_json(msg=msg)
        if os.path.isdir(file_path):
            msg = "Recursive directory not supported  %s " % (file_path)
            module.fail_json(msg=msg)


def get_connection(auth_args):
    """ separate function for get_connection ,might change ::
        still have figureout how to access env var defaults for auth """
    con = connection.Connection(**auth_args)
    return con


def create_stack(stack_name, template, wait, parameters, auth_args):
    con = get_connection(auth_args)
    resp = con.orchestration.find_stack(stack_name)
    if not (resp is None):
        resp = resp.to_dict()
        resp['changed'] = False
        return resp
    if wait == "yes":
        output = con.orchestration.create_stack(name=stack_name,
                                                parameters=parameters,
                                                template=template)
        resp = con.orchestration.wait_for_status(output,
                                                 status='CREATE_COMPLETE',
                                                 failures=['CREATE_FAILED'])
        resp = resp.to_dict()
        resp['changed'] = True
    else:
        resp = con.orchestration.create_stack(name=stack_name,
                                              parameters=parameters,
                                              template=template)
        resp = resp.to_dict()
        resp['changed'] = True
    return resp


def delete_stack(stack_name, wait, auth_args):
    con = get_connection(auth_args)
    con.orchestration.delete_stack(stack_name)
    resp = con.orchestration.find_stack(stack_name)
    if resp is None:
        resp = {}
        resp['changed'] = True
        return resp
    else:
        resp = {}
        resp['changed'] = True
        return resp


def main():
    module = AnsibleModule(
        argument_spec=dict(
            stack_name=dict(required=True, aliases=['name']),
            os_username=dict(required=False, aliases=['username']),
            os_password=dict(required=False, aliases=['password']),
            os_tenant_name=dict(required=False, aliases=['tenantname']),
            os_auth_url=dict(required=False, aliases=['url']),
            template=dict(required=True),
            state=dict(required=True),
            wait=dict(required=False, choices=['yes', 'no']),
            parameters=dict(required=False)
        ),
        required_one_of=[],
        supports_check_mode=True
    )

    stack_name = module.params['stack_name']
    state = module.params['state']
    template_path = os.path.expanduser(module.params['template'])
    check_file_paths(module, template_path)
    template = open(template_path, "r").read()
    wait = module.params['wait'] if 'wait' in module.params else "yes"
    if 'parameters' in module.params:
        parameters = ast.literal_eval(module.params['parameters'])
    else:
        parameters = None
    auth_args = {}
    if 'os_username' in module.params:
        auth_args['username'] = module.params['os_username']
    else:
        auth_args['username'] = os.environ.get('OS_USERNAME')
    if 'os_password' in module.params:
        auth_args['password'] = module.params['os_password']
    else:
        auth_args['password'] = os.environ.get('OS_PASSWORD')
    if 'os_tenant_name' in module.params:
        auth_args['project_name'] = module.params['os_tenant_name']
    else:
        auth_args['project_name'] = os.environ.get('OS_TENANT_NAME')
    if 'os_auth_url' in module.params:
        auth_args['auth_url'] = module.params['os_auth_url']
    else:
        auth_args['auth_url'] = os.environ.get('OS_AUTH_URL')
    for key in auth_args:
        if auth_args[key] is None:
            module.fail_json(msg="Auth Parameters missing")
    if state == "present":
        resp = create_stack(stack_name, template, wait, parameters, auth_args)
        resp["par"] = parameters
        module.exit_json(changed=resp['changed'], output=resp)
    elif state == "absent":
        resp = delete_stack(stack_name, wait, auth_args)
        module.exit_json(changed=resp['changed'], output=resp)


main()





class State(object):

    # currently unimplemented
    VALID_SUB_STATES = []

    def __init__(self, state, sub_state=None, ctx=None):
        """
        Initialize the State object with the current state, sub_state,
        and context as needed

        :param state: valid state
        :param sub_state: valid sub_state (default: None)
        :param ctx: valid context object, useful for configuration options
        """

        # Not sure if we really need to validate this data
        # It seems unnecessary, but could be valuable.
        # Additionally, because the states bits have been reworked, this
        # sort of validation no longer works as designed.
        self.ctx = ctx
        self.state = state
        self.sub_state = sub_state

#        if self._validate_state(state):
#            self.state = state
#        else:
#            raise StateError("Invalid State mentioned")

#        if sub_state == None:
#            self.sub_state = sub_state
#        elif self._validate_sub_state(sub_state):
#            self.sub_state = sub_state
#        else:
#            raise StateError("Invalid SubState mentioned")

    def _validate_state(self, state):
        """
        Validates the state based on linchpin.conf

        :param state: state name
        """

        VALID_STATES = self.ctx.cfgs["states"].keys()
        return state in VALID_STATES

    def _validate_sub_state(self, sub_state):
        """
        Not implemented yet
        ideally should validated substates of state based on linchpin.conf
        """

        return sub_state in State.VALID_SUB_STATES

    def __repr__(self):
        """
        Representational override of state object
        """

        if self.sub_state:
            return "{0}::{1}".format(self.state, self.sub_state)
        else:
            return self.state



#!/usr/bin/env python


def get_provider_resources(topo_output, res_type):
    provider_resources = []
    for host in topo_output:
        if host['resource_type'] == res_type:
            provider_resources.append(host)
    return provider_resources


class FilterModule(object):
    ''' A filter to fix network format '''
    def filters(self):
        return {
            'get_provider_resources': get_provider_resources
        }



"""
Inventory Providers List
"""
from linchpin.exceptions import LinchpinError

from .AWSInventory import AWSInventory
from .BeakerInventory import BeakerInventory
from .DuffyInventory import DuffyInventory
from .DummyInventory import DummyInventory
from .GCloudInventory import GCloudInventory
from .LibvirtInventory import LibvirtInventory
from .OpenstackInventory import OpenstackInventory
from .OvirtInventory import OvirtInventory

from .CFGInventoryFormatter import CFGInventoryFormatter
from .JSONInventoryFormatter import JSONInventoryFormatter

filter_classes = {
    "aws_ec2_res": AWSInventory,
    "beaker_res": BeakerInventory,
    "duffy_res": DuffyInventory,
    "dummy_res": DummyInventory,
    "nummy_res": DummyInventory,
    "gcloud_gce_res": GCloudInventory,
    "libvirt_res": LibvirtInventory,
    "os_server_res": OpenstackInventory,
    "ovirt_vms_res": OvirtInventory,
}

formatter_classes = {
    "cfg": CFGInventoryFormatter,
    "ini": CFGInventoryFormatter,
    "json": JSONInventoryFormatter
}


def get_driver(provider):

    if provider not in filter_classes:
            raise LinchpinError("Key {0} not found in"
                                " inventory provider dict".format(provider))

    return filter_classes[provider]


def get_all_drivers():
    return filter_classes


def get_inv_formatter(inv_type):
    if inv_type not in formatter_classes:
            raise LinchpinError("Key {0} not found in"
                                " formatters".format(inv_type))
    return formatter_classes[inv_type]


def get_all_inv_formatters():
    return formatter_classes



#!/user/bin/env python

# flake8: noqa

import os
import json
import yaml

from nose.tools import *
from six import iteritems

from linchpin.InventoryFilters import CFGInventoryFormatter

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser


def setup_json_inventory_formatter():
    global formatter
    global inv

    formatter = CFGInventoryFormatter.CFGInventoryFormatter()
    subdir = 'layouts'
    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', subdir)

    template = 'parsed-layout.json'
    template_file = open(mock_path+'/'+template)
    inv = json.load(template_file)


def setup_cfg_config():
    global config

    provider = 'general'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    cfg = 'config.yml'
    cfg_file = open(mock_path+'/'+cfg)
    config = yaml.load(cfg_file)
    cfg_file.close()


@with_setup(setup_json_inventory_formatter)
def test_add_sections():
    """ 
    should add a list of sections to the config, as well as "all"
    """
    section_list = ['foo', 'bar']

    formatter.add_sections(section_list)
    sections = formatter.config.sections()
    assert_equal(set(sections), set(['foo','bar','all']))


@with_setup(setup_json_inventory_formatter)
def test_set_children():
    """
    Should set the children of each inventory host group in the config
    """
    empty_inv = dict()
    formatter.set_children(empty_inv)

    formatter.set_children(inv)

    host_group='OSEv3'
    # if one if these does not exist, a NoOptionError will be thrown
    for child in inv['host_groups'][host_group]['children']:
        formatter.config.get('{0}:children'.format(host_group), child)


@with_setup(setup_json_inventory_formatter)
def test_set_vars():
    """
    Should add the vars in each host_group to the config
    """
    empty_inv = dict()
    formatter.set_vars(empty_inv)

    formatter.add_sections(inv['host_groups'].keys())
    formatter.set_vars(inv)

    host_group='OSEv3'
    for key, val in iteritems(inv['host_groups'][host_group]['vars']):
        var = formatter.config.get("{0}:vars".format(host_group), key)
        assert_equal(var, str(val))


@with_setup(setup_json_inventory_formatter)
def test_add_ips_to_groups():
    """
    """
    inven_hosts = inv['hosts']
    formatter.add_sections(inv['host_groups'].keys())
    formatter.add_ips_to_groups(inven_hosts, inv)
 

@with_setup(setup_json_inventory_formatter)
@with_setup(setup_cfg_config)
def test_add_common_vars():
    """
    """
    host_groups = inv['host_groups'].keys()
    formatter.add_sections(host_groups)
    formatter.set_vars(inv)
    formatter.add_common_vars(host_groups, inv, config)


@with_setup(setup_json_inventory_formatter)
def test_generate_inventory():
    """
    """
    inv_cfg = formatter.generate_inventory()
    # if the json is not valid, this will throw a ValueError and the test
    # will fail
    parser = ConfigParser()
    data = parser.readfp(inv_cfg)


import os
import sys
import logging

import xml.dom.minidom
import xml.etree.ElementTree as eT

from bkr.client import BeakerCommand, BeakerWorkflow, BeakerJob
from bkr.client import BeakerRecipeSet, BeakerRecipe
from bkr.common.hub import HubProxy
from bkr.common.pyconfig import PyConfigParser

from ansible.module_utils.basic import AnsibleModule

# WANT_JSON
# We want JSON input from Ansible. Please give it to us

LOG = logging.getLogger(__name__)
BEAKER_CONF = \
    (os.environ.get('BEAKER_CONF', '/etc/beaker/client.conf'))


class BkrConn(object):
    """
        Make connection to Beaker
    """
    enabled = True
    doc = xml.dom.minidom.Document()

    def __init__(self, logger=None, conf=None, **kwargs):
        self.conf = PyConfigParser()
        default_config = os.path.expanduser(BEAKER_CONF)
        self.conf.load_from_file(default_config)
        self.hub = HubProxy(logger=logger, conf=self.conf, **kwargs)


class BkrFactory(BkrConn):
    """
        Simple way to provision a job to the Beaker scheduler
    """
    def __init__(self, *args, **kwargs):
        super(BkrFactory, self).__init__(*args, **kwargs)

    def provision(self, *args, **kwargs):
        """
            provision resources in Beaker
        """
        # Break down kwargs for debug, dryrun, recipesets, and whiteboard
        debug = kwargs.get("debug", False)
        dryrun = kwargs.get("dryrun", False)
        recipesets = kwargs.get("recipesets", [])

        # Create Job
        job = BeakerJob(*args, **kwargs)

        # Add All Host Requirements
        for recipeset in recipesets:
            kwargs = self.create_recipesets(recipeset, **kwargs)
            family = kwargs.get("family", None)
            distro = kwargs.get("distro", None)
            task_params = kwargs.get("taskparam", [])
            # tasks are list of dictionaries which follows format
            # for beaker in a box
            # tasks = [ {arches:[], 'name': '/distribution/utils/dummy'}]
            # for beaker production
            # tasks = [ {arches:[], 'name': '/distribution/dummy'}]
            tasks = kwargs.get("tasks",
                               [{'arches': [], 'name': '/distribution/dummy'}])
            arch = kwargs.get("arch", "x86_64")
            ks_meta = kwargs.get("ks_meta", "")
            method = kwargs.get("method", "nfs")
            priority = kwargs.get("priority", "Normal")
            hostrequires = kwargs.get("hostrequires", [])
            reserve_duration = kwargs.get("reserve_duration", None)
            if reserve_duration:
                kwargs.update({"reserve_duration": "%s" % reserve_duration})
            tags = kwargs.get("tags", [])
            if tags:
                kwargs.update({"tag": tags})
            repos = kwargs.get("repos", [])
            baseurls = []
            for repo in repos:
                if "baseurl" in repo:
                    baseurls.append(repo.get("baseurl"))
            else:
                kwargs.update({"repo": baseurls})
            ks_append = kwargs.get("ks_append", [])
            ssh_key = kwargs.get("ssh_key", [])
            if ssh_key:
                ks_append.append("""%%post
mkdir -p /root/.ssh
cat >>/root/.ssh/authorized_keys << "__EOF__"
%s
__EOF__
restorecon -R /root/.ssh
chmod go-w /root /root/.ssh /root/.ssh/authorized_keys
%%end""" % '\n'.join(ssh_key))
                kwargs.update({"ks_append": ks_append})

            requested_tasks = []

            # adding arches=[] to every task definition
            for task in tasks:
                if not('arches' in task.keys()):
                    task['arches'] = []
                requested_tasks.append(task)

            # Tasks and harnesses
            if 'harness' in ks_meta:
                # Disable report plugins
                task_params.append("RSTRNT_DISABLED=01_dmesg_check "
                                   "10_avc_check")

                # Reserve the system after its installed
                kwargs.update({"reserve": True})
            else:
                requested_tasks.append({'arches': [],
                                        'name': '/distribution/reservesys'})

            # Update defaults
            kwargs.update({"suppress_install_task": True})
            kwargs.update({"method": method})
            kwargs.update({"priority": priority})

            # Must have family or distro
            if not family and not distro and not arch:
                sys.stderr.write("No Family or Distro and arch specified\n")
                sys.exit(1)

            if not requested_tasks:
                sys.stderr.write("You must specify a task to run\n")
                sys.exit(1)

            # Create Workflow
            wrkflow = BeakerWorkflow('BeakerWorkflow')

            # Create Base Recipe
            recipe_template = BeakerRecipe(*args, **kwargs)

            # Add Host Requirements
            for requirement in hostrequires:
                if 'force' in requirement:
                    # hostRequires element is created by BeakerRecipe, use it
                    hostrequires_node = recipe_template.node.getElementsByTagName('hostRequires')[0]  # noqa E501
                    # all other filters are ignored if the hostname is forced,
                    # so the use of 'force' is mutually exclusive with the use
                    # of any other 'hostRequires' filters
                    hostrequires_node.setAttribute('force',
                                                   requirement['force'])
                elif 'rawxml' in requirement:
                    requirement_node = xml.dom.minidom.parseString(
                        requirement['rawxml']).documentElement
                    recipe_template.addHostRequires(requirement_node)
                else:
                    # If force is not used, a requirement can be any number
                    # of differently formatted XML elements, each with their
                    # own combination of element name and valid attrs. So,
                    # the best we can do is generate XML based on the input,
                    # and only the "tag" key is required.
                    tag_name = requirement['tag']
                    requirement_node = self.doc.createElement(tag_name)
                    for attr, value in requirement.items():
                        # Force all values to str, which the XML writer expects.
                        requirement_node.setAttribute(attr, str(value))
                    # use the BeakerRecipe API to add the element
                    recipe_template.addHostRequires(requirement_node)

            # Add Distro Requirements
            recipe_template.addBaseRequires(*args, **kwargs)
            arch_node = self.doc.createElement('distro_arch')
            arch_node.setAttribute('op', '=')
            arch_node.setAttribute('value', arch)
            recipe_set = BeakerRecipeSet(**kwargs)
            if wrkflow.multi_host:
                for i in range(self.n_servers):
                    recipe_set.addRecipe(wrkflow.processTemplate(
                        recipe_template,
                        requested_tasks,
                        taskParams=task_params,
                        distroRequires=arch_node,
                        role='SERVERS',
                        **kwargs))
                for i in range(self.n_clients):
                    recipe_set.addRecipe(wrkflow.processTemplate(
                        recipe_template,
                        requested_tasks,
                        taskParams=task_params,
                        distroRequires=arch_node,
                        role='CLIENTS',
                        **kwargs))
            else:
                recipe_set.addRecipe(wrkflow.processTemplate(
                    recipe_template,
                    requested_tasks,
                    taskParams=task_params,
                    distroRequires=arch_node,
                    **kwargs))
            job.addRecipeSet(recipe_set)

        # jobxml
        jobxml = job.toxml(**kwargs)

        if debug:
            LOG.debug(jobxml)

        self.submitted_jobs = []

        if not dryrun:
            self.submitted_jobs.append(self.hub.jobs.upload(jobxml))
            LOG.info("Submitted: %s" % self.submitted_jobs)
        return self.submitted_jobs

    def create_recipesets(self, recipeset, **kwargs):
        kwargs = {}
        for key, values in recipeset.iteritems():
            kwargs.update({key: values})
        return kwargs

    def check_jobs(self, jobs):
        # slightly modified copy of the same method in bkr_info, used here to
        # let bkr_server return beaker job info in the same structure as
        # bkr_info so we can stash provisioning details in rundb before the
        # bkr_info task blocks on waiting for the beaker job(s) to finish.
        # Importing code from bkr_info is not an option due to ansible's
        # sandboxing; bkr_info is not guaranteed to be importable at runtime.
        resources = []
        bkrcmd = BeakerCommand('BeakerCommand')
        bkrcmd.check_taskspec_args(jobs)
        for task in jobs:
            myxml = self.hub.taskactions.to_xml(task)
            myxml = myxml.encode('utf8')
            root = eT.fromstring(myxml)
            # Using getiterator() since its backward compatible with py26
            for recipe in root.getiterator('recipe'):
                resources.append({'family': recipe.get('family'),
                                  'distro': recipe.get('distro'),
                                  'arch': recipe.get('arch'),
                                  'variant': recipe.get('variant'),
                                  'system': recipe.get('system'),
                                  'status': recipe.get('status'),
                                  'result': recipe.get('result'),
                                  'rid': recipe.get('id'),
                                  'id': recipe.get('job_id')})
        return resources

    def cancel_jobs(self, jobs, msg):
        """
            Cancel job in Beaker
        """
        for task in jobs:
            self.hub.taskactions.stop(task, 'cancel', msg)
            LOG.info("Cancel job %s in Beaker" % task)

    def get_url(self, bkr_id):
        """
        Constructs the Beaker URL for the job related to the provided Beaker
        ID. That ID should be all numeric, unless the structure of Beaker
        changes in the future. If that's the case, then the ID should be
        appropriately URL encoded to be appended to the end of a URL properly.
        """
        base = self.conf.get('HUB_URL', '')
        if base == '':
            raise Exception("Unable to construct URL")
        if base[-1] != '/':
            base += '/'
        return base + 'jobs/' + bkr_id


def main():
    module = AnsibleModule(argument_spec={
        'whiteboard': {'required': False, 'type': 'str'},
        'recipesets': {'required': True, 'type': 'list'},
        'job_group': {'default': '', 'type': 'str'},
        'state': {'default': 'present', 'choices': ['present', 'absent']},
        'cancel_message': {'default': 'Job canceled by LinchPin'},
        'max_attempts': {'default': 60, 'type': 'int'},
        'attempt_wait_time': {'default': 60, 'type': 'int'},
    })
    params = type('Args', (object,), module.params)
    factory = BkrFactory()
    recipesets = []
    for recipeset in params.recipesets:
        for x in range(0, recipeset.get('count', 1)):
            recipesets.append(recipeset)
    parsed_ids = {}
    if params.state == 'present':
        extra_params = {}
        if params.job_group:
            extra_params['job_group'] = params.job_group
        # Make provision
        try:
            job_ids = factory.provision(debug=True,
                                        recipesets=recipesets,
                                        whiteboard=params.whiteboard,
                                        **extra_params)
        except Exception as ex:
            module.fail_json(msg=str(ex))

        for job_id in job_ids:
            parsed_id = job_id[2:]
            parsed_ids[parsed_id] = factory.get_url(parsed_id)
        hosts = factory.check_jobs(job_ids)
        # pass out the provisioned job ids (or empty dict, if none) and
        # provisioning params to be used in the bkr_info task later
        module.exit_json(changed=True, ids=parsed_ids, hosts=hosts,
                         provision_params=module.params)
    else:  # state == absent, cancel provisioned jobs
        for recipeset in params.recipesets:
            # recipeset 'ids" value is set in the teardown playbook for use here
            factory.cancel_jobs(recipeset['ids'], params.cancel_message)
        module.exit_json(changed=True)


main()



import os
import sys
import json
import yaml

from nose.tools import assert_true
from nose.tools import with_setup

from linchpin import LinchpinAPI
from linchpin.utils.dataparser import DataParser
from linchpin.context import LinchpinContext
from linchpin.rundb import RunDB

from linchpin.validator import AnyofValidator

from linchpin.tests.mockdata.contextdata import ContextData


def setup_validator():

    """
    Perform setup of AnyofValidator, definitions, field, and value
    """

    global validator
    global topo

    lpc = LinchpinContext()
    lpc.load_config()
    lpc.load_global_evars()
    lpa = LinchpinAPI(lpc)

    schema_file = 'schema.json'
    base_path = '{0}'.format(os.path.dirname(os.path.realpath(__file__)))\
        .rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    sp = '{0}/{1}/{2}'.format(lib_path, 'mockdata/general', schema_file)
    schema = json.load(open(sp))

    validator = AnyofValidator(schema)

    topo_name = 'libvirt-new.yml'
    topo_file = '{0}/{1}/{2}'.format(lib_path, 'mockdata/libvirt', topo_name)
    topo = yaml.load(open(topo_file))



@with_setup(setup_validator)
def test_validate_anyof():
    document = { 'res_defs': topo['resource_groups'][0]['resource_definitions']}
    success = validator.validate(document)
    assert_true(success)


def main():
    pass


if __name__ == '__main__':
    sys.exit(main())



import os
import subprocess
import tempfile
import requests

from .fetch import Fetch
from linchpin.exceptions import LinchpinError


class FetchHttp(Fetch):

    def __init__(self, ctx, fetch_type, src, dest, cache_dir, root='',
                 root_ws='', ref=None):
        super(FetchHttp, self).__init__(ctx, fetch_type, dest, root=root,
                                        root_ws=root_ws, ref=ref)

        self.cache_dir = os.path.join(cache_dir, "http")
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        try:
            r = requests.get(src)
            r.raise_for_status()
        except requests.exceptions.HTTPError:
            raise LinchpinError("An HTTP error occurred")
        except requests.exceptions.RequestException:
            raise LinchpinError("Could not connect to given URL")

        if r.status_code != 200:
            raise LinchpinError("The entered url is invalid")

        self.src = src.rstrip('/')


    def fetch_files(self):

        key = "{0}|{1}".format(self.dest.replace(':', ''),
                               self.src.replace(':', ''))

        fetch_dir = self.cfgs["http"].get(key, None)
        self.td = self.call_wget(fetch_dir)

        self.td_w_root = '{0}/{1}/'.format(self.td, self.root)

        if not fetch_dir:
            self.write_cfg("http", key, self.td)


    def call_wget(self, fetch_dir=None):

        src_w_root = '{0}/{1}'.format(self.src, self.root)
        tempdir = None

        # globs to reject
        rej = '*html*'


        if not fetch_dir:
            fetch_dir = tempfile.mkdtemp(prefix="http_", dir=self.cache_dir)

        wget_args = ['wget', '-r', '-np', '-nH', '-q', '--reject',
                     rej, src_w_root, '-P', fetch_dir]

        retval = subprocess.call(wget_args)



        if retval != 0:
            try:
                os.rmdir(tempdir)
            except OSError:
                pass
            raise LinchpinError('Unable to fetch files with the following'
                                ' command:\n{0}'.format(" ".join(wget_args)))
        return fetch_dir



#!/usr/bin/env python

from .InventoryFilter import InventoryFilter
from .InventoryProviders import get_all_drivers
from .InventoryProviders import get_inv_formatter


class GenericInventory(InventoryFilter):

    def __init__(self, inv_format="cfg"):
        InventoryFilter.__init__(self)
        self.filter_classes = get_all_drivers()
        self.inv_formatter = get_inv_formatter(inv_format)()

    def get_host_data(self, res_output, config):
        """
        """
        host_data = []
        for res_grp in res_output:
            inv_filter = res_grp['resource_type']
            data = self.filter_classes[inv_filter]()\
                .get_host_data(res_grp, config)
            host_data.append(data)
        return host_data

    def get_hosts_by_count(self, host_data, count):
        """
        currently this function gets all the ips/hostname according to the
        order in which inventories are specified. later can be modified
        to work with user input
        """

        all_hosts = []
        for host in host_data:
            all_hosts.extend(host.keys())
        return all_hosts[:count]

    def populate_config(self, host_dict, res_output, config):
        """
        """
        populated_config = {}
        for provider in host_dict.keys():
            for host in host_dict[provider].keys():
                populated_config[host] = {}
                for var in config[provider]:
                    value = self.filter_classes[provider]().\
                        get_field_values(res_output, var)
                    populated_config[host][var] = value
        return populated_config

    def get_inventory(self, res_output, layout, topology, config):
        # get all the topology host_ips
        host_data = self.get_host_data(res_output, config)
        # sort it based on topology
        # get the count of all layout hosts needed
        layout_host_count = self.get_layout_hosts(layout)
        # generate hosts list based on the layout host count
        inven_hosts = self.get_hosts_by_count(host_data,
                                              layout_host_count)
        # adding sections to respective host groups
        host_groups = self.get_layout_host_groups(layout)

        self.inv_formatter.add_sections(host_groups)
        # set children for each host group
        self.inv_formatter.set_children(layout)
        # set vars for each host group
        self.inv_formatter.set_vars(layout)
        # add ip addresses to each host
        self.inv_formatter.add_ips_to_groups(inven_hosts, layout)
        # add common vars to host_groups
        self.inv_formatter.add_common_vars(host_groups, layout, host_data)
        return self.inv_formatter.generate_inventory()



#!/usr/bin/env python


def transform_os_server_output(res_def_out):
    res_def_out = res_def_out.get("results", [])
    res_def = {}
    res_def["ids"] = []
    res_def["openstack"] = []
    res_def["servers"] = []
    for ele in res_def_out:
        res_def["ids"].append(ele.get("id"))
        res_def["openstack"].append(ele.get("openstack", {}))
        res_def["servers"].append(ele.get("server", {}))
    return res_def


class FilterModule(object):
    ''' A filter fix distiller '''
    def filters(self):
        return {
            'transform_os_server_output': transform_os_server_output
        }



#!/usr/bin/env python

from __future__ import print_function

import os
import sys
import ast
import json
import click

from collections import OrderedDict

from linchpin.cli import LinchpinCli
from linchpin.exceptions import LinchpinError
from linchpin.cli.context import LinchpinCliContext
from linchpin.shell.click_default_group import DefaultGroup
from linchpin.shell.mutually_exclusive import MutuallyExclusiveOption


pass_context = click.make_pass_decorator(LinchpinCliContext, ensure=True)
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def _handle_results(ctx, results, return_code):
    """
    Handle results from the RunDB output and Ansible API.
    Either as a return value (retval) when running with the
    ansible console enabled, or as a list of TaskResult
    objects, and a return value.

    If a target fails along the way, this method immediately exits with the
    appropriate return value (retval). If the ansible console is disabled, an
    error message will be printed before exiting.

    :param results:
        The dictionary of results for each target.
    """

    output = 'Nothing to do. Check input and try again.'
    rcs = [return_code]

    for k, v in results.iteritems():

        output = '\nID: {0}\n'.format(k)
        output += 'Action: {0}\n'.format(v['action'])
        output += '\n{0:<20}\t{1:>6}\t{2:<5}\t{3:<10}\n'.format('Target',
                                                                'Run ID',
                                                                'uHash',
                                                                'Exit Code')
        output += '-------------------------------------------------\n'
        for target, run_data in v['summary_data'].iteritems():
            for rundb_id, data in run_data.iteritems():

                output += '{0:<20}\t{1:>6}\t{2:>5}'.format(target,
                                                           rundb_id,
                                                           data['uhash'])
                output += '\t{0:>9}\n'.format(data['rc'])

                # add all return codes to a little dictionary for use with
                # a future flag that allows to record failures, but continue
                # on anyway.
                rcs.append(data['rc'])

        task_results = v.get('results_data')
        if task_results:
            for target, results in task_results.iteritems():
                if results['task_results']:
                    tasks = results['task_results'][0].get('failed')

                    if not isinstance(tasks, int) and len(tasks):
                        trs = tasks

                        if trs is not None:
                            trs.reverse()
                            tr = trs[0]
                            msg = tr._check_key('msg')
                            ctx.log_state("\n--------------------------------"
                                          "----")
                            ctx.log_state("Task '{0}' failed with"
                                          " error '{1}' for Target:"
                                          " {2}".format(tr.task_name,
                                                        msg,
                                                        target))
                            ctx.log_state("\n\nAdd -vvvv to the linchpin"
                                          " command to see the stack trace")
                            ctx.log_state("----------------------------------"
                                          "--\n")

                # FIXME make sure the return_code is valid here

    # PRINT OUTPUT RESULTS HERE
    ctx.log_state(output)
    uar = lpcli.get_cfg('lp', 'use_actual_rcs')
    use_actual_rcs = ast.literal_eval(uar.title())

    if use_actual_rcs:
        return_code = min(rcs)
    else:
        return_code = sum(rc for rc in rcs)

    sys.exit(return_code)


@click.group(cls=DefaultGroup, default_if_no_args=True, default='help',
             invoke_without_command=True,
             context_settings=CONTEXT_SETTINGS)
@click.option('-c', '--config', type=click.Path(), envvar='LP_CONFIG',
              help='Path to config file')
@click.option('-p', '--pinfile', envvar='PINFILE', metavar='PINFILE',
              help='Use a name for the PinFile different from'
                   ' the configuration.')
@click.option('-d', '--template-data', metavar='TEMPLATE_DATA',
              help="Template data passed to PinFile template.\n\n"
                   "If template data is from a file, it must be"
                   " prepended with an '@' character."
              )
@click.option('-o', '--output-file', '--output-pinfile', 'outfile',
              metavar='OUTPUT_FILE',
              help='Write out PinFile to provided location')
@click.option('-w', '--workspace', type=click.Path(), envvar='WORKSPACE',
              help='Use the specified workspace. Also works if the'
                   ' familiar Jenkins WORKSPACE environment variable is set')
@click.option('-v', '--verbose', count=True, default=0,
              help='Enable verbose output')
@click.option('--version', is_flag=True,
              help='Prints the version and exits')
@click.option('--creds-path', type=click.Path(), envvar='CREDS_PATH',
              help='Use the specified credentials path. Also works'
                   ' if CREDS_PATH environment variable is set')
@click.option('--ask-vault-pass', is_flag=True, default=False,
              help='Prompts for vault password')
@pass_context
def runcli(ctx, config, pinfile, template_data, outfile,
           workspace, verbose, version, creds_path, ask_vault_pass):
    """linchpin: hybrid cloud orchestration"""

    ctx.load_config(lpconfig=config)
    # workspace arg in load_config used to extend linchpin.conf
    ctx.load_global_evars()
    ctx.setup_logging()

    ctx.verbosity = verbose

    # if the pinfile is a template, data will be passed here
    ctx.pf_data = template_data
    ctx.outfile = None
    if outfile:
        ctx.outfile = outfile
        ctx.set_cfg('tmp', 'outfile', outfile)

    ctx.pinfile = None
    if pinfile:
        ctx.pinfile = pinfile

    if version:
        ctx.log_state('linchpin version {0}'.format(ctx.version))
        sys.exit(0)

    if creds_path is not None:
        ctx.set_evar('creds_path',
                     os.path.realpath(os.path.expanduser(creds_path)))

    if workspace is not None:
        ctx.workspace = os.path.realpath(os.path.expanduser(workspace))
        ctx.log_debug("ctx.workspace: {0}".format(ctx.workspace))

    ctx.ask_vault_pass = ask_vault_pass


    # global LinchpinCli placeholder
    global lpcli
    lpcli = LinchpinCli(ctx)


@runcli.command('help', short_help='Prints help')
@click.pass_context
def help(ctx):
    """
    Print help
    """

    print(ctx.parent.get_help())


@runcli.command('init', short_help='Initializes a linchpin project.')
@click.argument('provider', metavar='PROVIDER', required=False, nargs=1)
@pass_context
def init(ctx, provider):
    """
    Initializes a linchpin project, which generates an example PinFile, and
    creates the necessary directory structure for topologies and layouts.

    Utilizes lp_fetch with the following parameters:

    remote:         git://github.com/CentOS-PaaS-SIG/linchpin
    root:           workspaces
    fetch_type:     workspace
    fetch_protocol: FetchGit
    fetch_ref:      master
    dest_ws:        Unused, will use workspace/provider
    nocache:        True


    """

    remote = ctx.get_cfg('init', 'remote',
                         default='git://github.com/CentOS-PaaS-SIG/linchpin')
    root = ctx.get_cfg('init', 'root', default='workspaces/dummy')
    fetch_type = ctx.get_cfg('init', 'fetch_type', default='workspace')
    fetch_proto = 'FetchGit'
    fetch_ref = ctx.get_cfg('init', 'fetch_ref', default='master')
    nocache = ast.literal_eval(ctx.get_cfg('init', 'nocache', default='True'))

    if provider:
        root = 'workspaces/{0}'.format(provider)

    try:
        lpcli.lp_fetch(remote, root=root, fetch_type=fetch_type,
                       fetch_protocol=fetch_proto, fetch_ref=fetch_ref,
                       dest_ws=None, nocache=nocache)
    except LinchpinError as e:
        ctx.log_state(e)
        sys.exit(1)


@runcli.command()
@click.argument('targets', metavar='TARGETS', required=False,
                nargs=-1)
@click.option('-r', '--run-id', metavar='run_id', type=int,
              help='Idempotently provision using `run-id` data',
              cls=MutuallyExclusiveOption, mutually_exclusive=["tx_id"])
@click.option('-t', '--tx-id', metavar='tx_id', type=int,
              help='Provision resources using the Transaction ID (tx-id)',
              cls=MutuallyExclusiveOption, mutually_exclusive=["run_id"])
@click.option('--inventory-format', '--if', 'inventory_format', default="cfg",
              metavar='INVENTORY_FORMAT', help="Inventory format cfg/json")
@click.option('--ignore-failed-hooks', '--ifh', 'ignore_failed_hooks',
              is_flag=True, default=False, metavar='IGNORE_FAILED_HOOKS',
              help='Ignores failed hooks')
@click.option('--no-hooks', '--nh', 'no_hooks', is_flag=True, default=None,
              metavar='NO_HOOKS', help='Do not run hooks')
@pass_context
def up(ctx, targets, run_id, tx_id, inventory_format, ignore_failed_hooks,
       no_hooks):
    """
    Provisions nodes from the given target(s) in the given PinFile.

    The `run_id` requires an associated target, where the `tx_id` will look up
    the targets from the specified transaction.

    The data from the targets is obtained from the PinFile (default).
    By setting `use_rundb_for_actions = True` in linchpin.conf, any
    up transaction which use the `-r/--run_id` or `-t/--tx_id` option will
    obtain target data from the RunDB.

    targets:    Provision ONLY the listed target(s). If omitted, ALL targets in
    the appropriate PinFile will be provisioned.

    run-id:     Use the data from the provided run_id value
    """
    vault_pass = os.environ.get('VAULT_PASSWORD', '')

    if ctx.ask_vault_pass:
        vault_pass = click.prompt("enter vault password", hide_input=True)

    ctx.set_evar('vault_pass', vault_pass)

    if ignore_failed_hooks:
        ctx.set_cfg("hook_flags", "ignore_failed_hooks", ignore_failed_hooks)
    if no_hooks:
        ctx.set_cfg("hook_flags", "no_hooks", no_hooks)

    if tx_id:
        try:
            return_code, results = lpcli.lp_up(targets=targets,
                                               tx_id=tx_id,
                                               inv_f=inventory_format)
            _handle_results(ctx, results, return_code)
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)
    else:  # if tx_id is not passed, use run_id as a baseline
        if (not len(targets) or len(targets) > 1) and run_id:
            raise click.UsageError("A single target is required when calling"
                                   " destroy with `--run_id` option")
        try:
            return_code, results = lpcli.lp_up(targets=targets,
                                               run_id=run_id,
                                               tx_id=tx_id,
                                               inv_f=inventory_format)
            _handle_results(ctx, results, return_code)
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)

#    try:
#        return_code, results = lpcli.lp_up(targets=targets, run_id=run_id)
#        _handle_results(ctx, results, return_code)
#    except LinchpinError as e:
#        ctx.log_state(e)
#        sys.exit(1)


@runcli.command()
@click.argument('targets', metavar='TARGET', required=False,
                nargs=-1)
@click.option('-r', '--run-id', metavar='run_id', type=int,
              help='Destroy resources using a target-based ID (run-id)',
              cls=MutuallyExclusiveOption, mutually_exclusive=["tx_id"])
@click.option('-t', '--tx-id', metavar='tx_id', type=int,
              help='Destroy resources using the transaction ID (tx-id)',
              cls=MutuallyExclusiveOption, mutually_exclusive=["run_id"])
@click.option('--ignore-failed-hooks', '--ifh', 'ignore_failed_hooks',
              metavar='ignore_failed_hooks', is_flag=True, default=False,
              help='Ignores failed hooks')
@click.option('--no-hooks', '--nh', 'no_hooks', metavar='NO_HOOKS',
              is_flag=True, help='Do not run hooks')
@pass_context
def destroy(ctx, targets, run_id, tx_id, ignore_failed_hooks, no_hooks):
    """
    Destroys resources using either the run_id or tx_id (mutually exclusive).

    The run_id requires an associated target, where the tx_id will look up
    the targets from the specified transaction.

    The data from the targets is obtained from the PinFile (default) or from
    the RunDB (by setting `use_rundb_for_actions = True` in linchpin.conf).

    targets:    Destroy ONLY the listed target(s). If omitted, ALL targets in
    the appropriate PinFile will be destroyed.

    """
    vault_pass = os.environ.get('VAULT_PASSWORD', '')
    if ctx.ask_vault_pass:
        vault_pass = click.prompt("enter vault password", hide_input=True)

    ctx.set_evar('vault_pass', vault_pass)

    if ignore_failed_hooks:
        ctx.set_cfg("hook_flags", "ignore_failed_hooks", ignore_failed_hooks)
    if no_hooks:
        ctx.set_cfg("hook_flags", "no_hooks", no_hooks)


    if tx_id:
        try:
            return_code, results = lpcli.lp_destroy(targets=targets,
                                                    tx_id=tx_id)
            _handle_results(ctx, results, return_code)
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)
    else:  # if tx_id is not passed, use run_id as a baseline
        if (not len(targets) or len(targets) > 1) and run_id:
            raise click.UsageError("A single target is required when calling"
                                   " destroy with `--run_id` option")
        try:
            return_code, results = lpcli.lp_destroy(targets=targets,
                                                    run_id=run_id,
                                                    tx_id=tx_id)
            _handle_results(ctx, results, return_code)
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)


@runcli.command()
@click.argument('remote', default=None, required=True, nargs=1)
@click.option('-t', '--type', 'fetch_type', metavar='TYPE', required=False,
              default='workspace', help='Which component of a workspace to'
              ' fetch. (Default: workspace)')
@click.option('-r', '--root', metavar='ROOT', default='',
              help='Use this to specify the location of the workspace'
                   ' within the root url. If root is not set, the root'
                   ' of the given remote will be used.')
@click.option('--dest', 'dest_ws', metavar='DEST', default=None,
              help='Workspaces destination, the fetched workspace will be'
                   ' relative to this location. (Overrides -w/--workspace)')
@click.option('--branch', 'fetch_ref', metavar='REF', default=None,
              help='Specify the git branch. Used only with'
                   ' git protocol (eg. master).')
@click.option('--git', 'fetch_protocol', flag_value='FetchGit', default=True,
              help='Remote is a git repository (default)')
@click.option('--web', 'fetch_protocol', flag_value='FetchHttp',
              help='Remote is a web directory')
@click.option('--nocache', is_flag=True,
              help='Do not check the cached time, just copy the data to the'
                   ' destination')
@pass_context
def fetch(ctx, remote, fetch_type, root, dest_ws,
          fetch_ref, fetch_protocol, nocache):
    """
    Fetches a specified linchpin workspace or component from a remote location

    """

    if not fetch_type:
        fetch_type = 'workspace'

    try:
        lpcli.lp_fetch(remote, root=root, fetch_type=fetch_type,
                       fetch_protocol=fetch_protocol, fetch_ref=fetch_ref,
                       dest_ws=dest_ws, nocache=nocache)
    except LinchpinError as e:
        ctx.log_state(e)
        sys.exit(1)


@runcli.command()
@click.argument('targets', metavar='TARGETS',
                required=False, default=None, nargs=-1)
@click.option('--view', metavar='VIEW', default='target', required=False,
              help='Type of view display (default: target)')
@click.option('-c', '--count', metavar='COUNT', default=3, required=False,
              help='(up to) number of records to return (default: 3)')
@click.option('-f', '--fields', metavar='FIELDS', required=False,
              help='List the fields to display')
@click.option('-t', '--tx-id', multiple=True, type=int,
              metavar='TX_ID', required=False,
              help="Display a specific transaction by ID (tx_id)."
                   " Only works with '--view=tx'")
@click.option('--output-format', type=str, default="cfg",
              metavar='output_format', required=False,
              help="Inventory output format")
@click.option('--output-type', type=str,
              metavar='output_type', required=False,
              help="default inventory")
@click.option('--target', metavar='target', default="all",
              help='If multiple targets are mentioned \
                    takes parameter for target to be used.\
                    By default all are displayed\
                    displayed')
@pass_context
def journal(ctx, targets, fields, count, view,
            tx_id, output_format, output_type, target):
    """
    Display information stored in Run Database

    view:       How the journal is displayed

                'target': show results of transactions on listed targets
                (or all if omitted)

                'tx': show results of each transaction, with results
                of associated targets used

    (Default: target)

    count:      Number of records to show per target

    targets:    Display data for the listed target(s). If omitted, the latest
                records for any/all targets in the RunDB will be displayed.

    fields:     Comma separated list of fields to show in the display. Used
    only with `--view=target`.
    (Default: action, uhash, rc. Additional fields: start, end)
    """

    if output_type == "inventory":
        inventories = lpcli._write_to_inventory(inv_path=ctx.outfile,
                                                inv_format=output_format)
        if target == "all":
            click.echo("By default all targets inventories are displayed to stdout\
                        For specific target please use --target option")
            for target in inventories:
                click.echo(inventories[target])
        else:
            try:
                click.echo(inventories[target])
            except IndexError as e:
                click.echo("Invalid target")
                click.echo(e.message)
        return inventories

    if view == 'target':

        try:
            journal = lpcli.lp_journal(view=view, targets=targets,
                                       fields=fields, count=count)
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)

        all_fields = json.loads(lpcli.get_cfg('lp', 'rundb_schema')).keys()

        if not fields:
            fields = ['action', 'uhash', 'rc']
        else:
            fields = fields.split(',')

        invalid_fields = [field for field in fields if field not in all_fields]

        if invalid_fields:
            ctx.log_state('The following fields passed in are not valid: {0}'
                          ' \nValid fields are {1}'.format(fields, all_fields))
            sys.exit(89)

        no_records = []
        output = 'run_id\t'

        for f in fields:
            output += '{0:>10}\t'.format(f)

        output += '\n'
        output += '--------------------------------------------------'


        if len(journal):
            for target, values in journal.iteritems():
                keys = values.keys()

                if len(keys):
                    print('\nTarget: {0}'.format(target), file=sys.stderr)
                    print(output, file=sys.stderr)
                    keys.sort(reverse=True)
                    for run_id in keys:
                        if int(run_id) > 0:
                            out = '{0:<7}\t'.format(run_id)
                            for f in fields:
                                out += '{0:>9}\t'.format(values[run_id][f])

                            print(out, file=sys.stderr)
                else:
                    no_records.append(target)


            if no_records:
                no_out = '\nNo records for targets:'

                for rec in no_records:
                    no_out += ' {0}'.format(rec)

                print(no_out, file=sys.stderr)

        else:
            print('No targets available for journal.'
                  ' Please provision something. :)', file=sys.stderr)
        print('\n')
        ctx.log_state(output)

    elif view == 'tx':

        try:
            j = lpcli.lp_journal(view=view, count=count, tx_ids=tx_id)
            if not len(tx_id):
                journal = OrderedDict(reversed(sorted(j.items())))
            else:
                journal = OrderedDict(j.items())
        except LinchpinError as e:
            ctx.log_state(e)
            sys.exit(1)

        output = ''

        if len(journal):
            for lp_id, v in journal.iteritems():

                if v:
                    output += '\nID: {0}\t\t\t'.format(lp_id)
                    output += 'Action: {0}\n'.format(v['action'])
                    output += '\n{0:<20}\t{1:>6}'.format('Target', 'Run ID')
                    output += '\t{0:<5}\t{1:<10}\n'.format('uHash', 'Exit Code')
                    output += '----------------------------------------------'
                    output += '---\n'

                    for targets in v['targets']:
                        for target, values in targets.iteritems():
                            for rundb_id, data in values.iteritems():
                                output += '{0:<20}\t{1:>6}\t'.format(target,
                                                                     rundb_id)
                                output += '{0:>5}'.format(data['uhash'])
                                output += '\t{0:>9}\n'.format(data['rc'])

                    output += '\n======================================='
                    output += '==========\n'

        else:
                output += '\n==================NO TRANSACTIONS======'
                output += '==========\n'

        ctx.log_state(output)


@runcli.command()
@click.option('--old-schema', '-o', is_flag=True)
@click.argument('targets', metavar='TARGETS', required=False,
                nargs=-1)
@pass_context
def validate(ctx, targets, old_schema):
    """
    Validate topologies for the given target(s) in the given PinFile.

    The data from the targets is obtained from the PinFile (default).

    targets:    Validate ONLY the listed target(s). If omitted, ALL targets in
    the appropriate PinFile will be validate

    """

    try:
        old_schema = False
        return_code, results = lpcli.lp_validate(targets=targets,
                                                 old_schema=old_schema)
        for target, item in results.iteritems():
            result = ""
            for kind, outcome in item.iteritems():
                if outcome == "valid" or outcome == "valid under old schema":
                    result += "[SUCCESS] {0} for target '{1}' is "\
                        "{2}\n".format(kind, target, outcome)
                else:
                    result += "[ERROR] {0} for target '{1}' does not "\
                              "validate\n{2}".format(kind, target, outcome)
            ctx.log_state(result)

        if old_schema:
            warning = """
Topologies valid under the old schema may have older directives.
It is suggested to update any of the following:
    res_group_type -> resource_group_type
    res_defs -> resource_definitions
    res_name -> name
    type -> role
    res_type -> role"""
            ctx.log_state(warning)
        sys.exit(return_code)
    except LinchpinError as e:
        ctx.log_state(e)
        sys.exit(1)


@runcli.command()
@click.argument('providers', metavar='PROVIDERS', required=False,
                nargs=-1)
@click.option('--ask-sudo-pass', is_flag=True, default=False,
              help='Prompts for sudo password for package installations')
@pass_context
def setup(ctx, providers, ask_sudo_pass):
    """
    Validate topologies for the given target(s) in the given PinFile.
    The data from the targets is obtained from the PinFile (default).
    targets:    Validate ONLY the listed target(s). If omitted, ALL targets in
    the appropriate PinFile will be validate
    """
    lpcli.ctx.set_evar("ask_sudo_pass", ask_sudo_pass)
    return_code, output = lpcli.lp_setup(providers)
    return sys.exit(return_code)


def main():
    # print("entrypoint")
    pass



#!/usr/bin/env python


def map_results(results, attr, subattr):
    output = []
    print("Inside filter")
    print(attr)
    print(subattr)
    print("results")
    print(results)
    for task_result in results:
        output.append(task_result[attr][subattr])
    print("Output")
    print(output)
    return output


class FilterModule(object):
    ''' A filter to fix network format '''
    def filters(self):
        return {
            'map_results': map_results
        }



#!/usr/bin/env python
import os

from time import sleep
from sys import stderr
import xml.etree.ElementTree as eT
from ansible.module_utils.basic import AnsibleModule

from bkr.client import BeakerCommand
from bkr.common.hub import HubProxy
from bkr.common.pyconfig import PyConfigParser


BEAKER_CONF = os.environ.get('BEAKER_CONF', '/etc/beaker/client.conf')


def _jprefix(job_ids):
    return ["J:" + _id for _id in job_ids]


class BeakerTargets(object):
    def __init__(self, params, logger=None):
        # params from AnsibleModule argument_spec below
        self.ids = params['ids']
        provision_params = params['provision_params']

        # Set wait methods from provision params, with reasonable defaults
        self.wait_time = provision_params.get('attempt_wait_time', 60)
        self.max_attempts = provision_params.get('max_attempts', 60)

        # set up beaker connection
        self.conf = PyConfigParser()
        default_config = os.path.expanduser(BEAKER_CONF)
        self.conf.load_from_file(default_config)
        self.hub = HubProxy(logger=logger, conf=self.conf)

    def get_system_statuses(self):
        """
        Checks on the status of a set of Beaker jobs (ids) and returns their
        hostname once the jobs have reached their defined status.
        """
        attempts = 0
        while attempts < self.max_attempts:
            job_results, all_count = self._check_jobs()
            pass_count = 0
            for resource in job_results:
                result = resource['result']
                status = resource['status']
                print >> stderr, "status: %s, result: %s" % (status, result)
                if status not in ['Cancelled', 'Aborted']:
                    if result == 'Pass':
                        pass_count += 1
                    elif result in ['Fail', 'Warn', 'Panic', 'Completed']:
                        raise Exception("System failed with state"
                                        " '{0}'".format(result))
                elif status == 'Aborted':
                    raise Exception("System aborted")
                elif status == 'Cancelled':
                    raise Exception("System canceled")
            attempts += 1
            if pass_count == all_count:
                return job_results
            sleep(self.wait_time)

        # max attempts exceeded, cancel jobs and fail
        for job_id in _jprefix(self.ids):
            self.hub.taskactions.stop(job_id, 'cancel',
                                      'Provision request timed out')
        # Fail with error msg, include results from last attempt to include
        # in topology outputs even if provisioning failed so a destroy still
        # cancels jobs
        msg = ("{0} system(s) never completed in {1} polling attempts, jobs "
               "have been cancelled: {2}".format(
                   all_count - pass_count, attempts, ', '.join(self.ids)))
        raise Exception(msg, job_results)

    def _check_jobs(self):
        """
            Get state of a job in Beaker
        """
        jobs = _jprefix(self.ids)
        resources = []
        bkrcmd = BeakerCommand('BeakerCommand')
        bkrcmd.check_taskspec_args(jobs)
        for task in jobs:
            myxml = self.hub.taskactions.to_xml(task)
            myxml = myxml.encode('utf8')
            root = eT.fromstring(myxml)
            # Using getiterator() since its backward compatible with py26
            for recipe in root.getiterator('recipe'):
                resources.append({'family': recipe.get('family'),
                                  'distro': recipe.get('distro'),
                                  'arch': recipe.get('arch'),
                                  'variant': recipe.get('variant'),
                                  'system': recipe.get('system'),
                                  'status': recipe.get('status'),
                                  'result': recipe.get('result'),
                                  'rid': recipe.get('id'),
                                  'id': recipe.get('job_id')})
        return resources, len(resources)


def main():
    mod = AnsibleModule(argument_spec={
        'ids': {'type': 'dict'},
        'provision_params': {'type': 'dict'},
    })
    beaker = BeakerTargets(mod.params)
    try:
        if len(beaker.ids) > 1:
            mod.warn('When using multiple resource_definitions for beaker '
                     'resources, only the provisioning parameters '
                     '(max_attempts, attempt_wait_time) from the first '
                     'resource_definition will be used. Consider using a '
                     'single resource_definition with multiple recipes or '
                     'recipesets instead.')
        results = beaker.get_system_statuses()
        mod.exit_json(hosts=results, changed=True)
    except Exception as ex:
        msg = ": For more details please check jobs on beaker"
        msg = str(ex) + msg
        mod.fail_json(msg=msg, changed=True)


# import module snippets
main()



#!/usr/bin/env python

# flake8: noqa

import os
import json
import yaml

from nose.tools import *

from linchpin.InventoryFilters import OvirtInventory

def setup_ovirt_inventory_filter():
    global filter
    global topo

    filter = OvirtInventory.OvirtInventory()

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    topology = 'linchpin.benchmark'
    topo_file = open(mock_path+'/'+topology)
    topo = json.load(topo_file)['17']['targets'][0]['general-inventory']['outputs']['resources']
    topo_file.close()


def setup_ovirt_config():
    global config

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    cfg = 'PinFile'
    cfg_file = open(mock_path+'/'+cfg)
    config = yaml.load(cfg_file)['general-inventory']['cfgs']
    cfg_file.close()


def setup_ovirt_layout():
    global layout

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    template = 'linchpin.benchmark'
    template_file = open(mock_path+'/'+template)
    layout = json.load(template_file)['17']['targets'][0]['general-inventory']['inputs']['layout_data']['inventory_layout']


@with_setup(setup_ovirt_inventory_filter)
@with_setup(setup_ovirt_config)
def test_get_host_data():
    """
    """
    host_data = filter.get_host_data(topo[8], config)
    expected_vars = ['__IP__']
    for host in host_data:
        assert_equal(set(host_data[host].keys()), set(expected_vars))


@with_setup(setup_ovirt_inventory_filter)
@with_setup(setup_ovirt_config)
def test_get_host_ips():
    """
    """
    host_data = filter.get_host_data(topo[8], config)
    ips = filter.get_host_ips(host_data)
    expected_hosts = ["155.132.55.17", "11.50.197.1"]
    assert_equal(set(host_data.keys()), set(expected_hosts))


@with_setup(setup_ovirt_inventory_filter)
@with_setup(setup_ovirt_config)
@with_setup(setup_ovirt_layout)
def test_get_inventory():
    """
    """
    inventory = filter.get_inventory(topo, layout, config)
    assert_true(inventory)



import os
import json
import yaml

from nose.tools import assert_true
from nose.tools import assert_equal
from nose.tools import with_setup

from linchpin import LinchpinAPI
from linchpin.context import LinchpinContext

from linchpin.validator import Validator


def setup_validator():

    """
    Perform setup of Validator,
    """

    global validator
    global pinfile

    lpc = LinchpinContext()
    lpc.load_config()
    lpc.load_global_evars()
    lpa = LinchpinAPI(lpc)
    validator = Validator(lpa.ctx, lpa.pb_path, lpa.pb_ext)

    schema_file = 'schema.json'
    base_path = '{0}'.format(os.path.dirname(os.path.realpath(__file__)))\
        .rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    sp = '{0}/{1}/{2}'.format(lib_path, '../provision/roles/dummy/files', schema_file)
    schema = json.load(open(sp))

    pf_name = 'PinFile-complete.yml'
    pf_file = '{0}/{1}/{2}'.format(lib_path, 'mockdata/dummy', pf_name)
    pinfile = yaml.load(open(pf_file))


def setup_old_topology():
    global old_topology

    pf_name = "PinFile-old.yml"
    base_path = '{0}'.format(os.path.dirname(os.path.realpath(__file__)))\
        .rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    pf_file = '{0}/{1}/{2}'.format(lib_path, 'mockdata/dummy', pf_name)
    old_topology = yaml.load(open(pf_file))['dummy-new']['topology']


@with_setup(setup_validator)
def test_validate():
    validator.validate(pinfile['dummy-new'])


@with_setup(setup_validator)
def test_validate_pretty():
    results = validator.validate_pretty(pinfile['dummy-new'], 'dummy-new')
    assert_true(results)


@with_setup(setup_validator)
def test_validate_topology():
    success = False
    topo_data = pinfile['dummy-new']['topology']
    try:
        resources = validator.validate_topology(topo_data)
        success = len(resources)
    except Exception:
        pass

    assert success


@with_setup(setup_validator)
def test_validate_layout():
    success = False
    layout_data = pinfile['dummy-new']['layout']

    validator.validate_layout(layout_data)


@with_setup(setup_validator)
def test_validate_resource_group():
    success = False
    res_grp = pinfile['dummy-new']['topology']['resource_groups'][0]

    # this will pass unless an exception is thrown
    validator.validate_resource_group(res_grp)


@with_setup(setup_validator)
def test_validate_cfgs():
    pass

@with_setup(setup_validator)
def test_find_playbook_path():
    pb_path = validator._find_playbook_path('dummy')

    assert os.path.exists(os.path.expanduser(pb_path))


def test_gen_error_msg():
    error = {'res_defs': [{0: [{u'name': ["field 'name' is required"]}]}]}

    error_msg = validator._gen_error_msg('', '', error)
    expected_msg = "res_defs[0][name]: field 'name' is required\n"

    assert_equal(error_msg, expected_msg)

@with_setup(setup_validator)
@with_setup(setup_old_topology)
def test_convert_topology():
    print old_topology
    validator._convert_topology(old_topology)
    assert_equal(old_topology, pinfile['dummy-new']['topology'])



#!/usr/bin/env python

# flake8: noqa

import os
import json

from nose.tools import *
from six import iteritems

import logging
from unittest import TestCase

from linchpin.InventoryFilters import GenericInventory

def setup_inventory_filter():
    global filter
    global inv

    filter = GenericInventory.GenericInventory()
    subdir = 'layouts'
    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', subdir)

    template = 'parsed-layout.json'
    template_file = open(mock_path+'/'+template)
    inv = json.load(template_file)

@with_setup(setup_inventory_filter)
def test_get_layout_hosts():
    """
    Should return the number of hosts in an inventory file.
    """
    hosts = filter.get_layout_hosts(inv)
    assert_equal(hosts, 3)

@with_setup(setup_inventory_filter)
def test_get_layout_host_groups():
    """
    Should return a list of the host_groups in an inventory file.
    """
    
    host_groups = filter.get_layout_host_groups(inv)
    # casting to sets ensures that only the elements (not order) are considered
    assert_equal(set(host_groups),
        set(['masters', 'etcd', 'nodes', 'nfs', 'glusterfs']))

@with_setup(setup_inventory_filter)
def test_add_sections():
    """
    should add a list of sections to the config, as well as "all"
    """
    section_list = ['foo', 'bar']

    filter.add_sections(section_list)
    sections = filter.config.sections()
    assert_equal(set(sections), set(['foo','bar','all']))

@with_setup(setup_inventory_filter)
def test_set_children():
    """
    Should set the children of each inventory host group in the config
    """
    empty_inv = dict()
    filter.set_children(empty_inv)

    filter.set_children(inv)

    host_group='OSEv3'
    # if one if these does not exist, a NoOptionError will be thrown
    for child in inv['host_groups'][host_group]['children']:
        filter.config.get('{0}:children'.format(host_group), child)

@with_setup(setup_inventory_filter)
def test_set_vars():
    """
    Should add the vars in each host_group to the config
    """
    empty_inv = dict()
    filter.set_vars(empty_inv)

    filter.set_vars(inv)

    host_group='OSEv3'
    for key, val in iteritems(inv['host_groups'][host_group]['vars']):
        var = filter.config.get("{0}:vars".format(host_group), key)
        assert_equal(var, str(val))

@with_setup(setup_inventory_filter)
def test_add_ips_to_groups():
    """
    I don't know what this is
    """
    inven_hosts = inv['hosts']
    filter.add_ips_to_groups(inven_hosts, inv)

@with_setup(setup_inventory_filter)
def test_add_common_vars():
    """
    I don't know what this is
    """
    host_groups = inv['host_groups'].keys()
    filter.add_sections(host_groups)
    filter.set_vars(inv)
    filter.add_ips_to_groups(host_groups, inv)
    filter.add_common_vars(host_groups, inv)


#!/usr/bin/env python
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from .InventoryFilter import InventoryFilter


class OpenstackInventory(InventoryFilter):
    DEFAULT_HOSTNAMES = ["accessIPv4", "public_v4", "private_v4"]

    def get_host_data(self, res, cfgs):
        host_data = {}
        if res['resource_type'] != 'os_server_res':
            return host_data
        var_data = cfgs.get('openstack', {})
        if var_data is None:
            var_data = {}

        if 'results' in res.keys():
            for result in res.get('results', []):
                if 'openstack' in result.keys():
                    os_vars = result.get('openstack', [])
                    host = self.get_hostname(os_vars, var_data,
                                             self.DEFAULT_HOSTNAMES)
                    hostname_var = host[0]
                    hostname = host[1]
                    host_data[hostname] = {}
                    if '__IP__' not in var_data.keys():
                        var_data['__IP__'] = hostname_var
                        host_data[hostname] = {}
                    self.set_config_values(host_data[hostname], os_vars,
                                           var_data)
                else:
                    continue
        else:
            grp = res.get('openstack', [])
            if isinstance(grp, list):
                for server in grp:
                    host = self.get_hostname(server, var_data,
                                             self.DEFAULT_HOSTNAMES)
                    hostname_var = host[0]
                    hostname = host[1]
                    host_data[hostname] = {}
                    if '__IP__' not in var_data.keys():
                        var_data['__IP__'] = hostname_var
                        host_data[hostname] = {}
                    host_data[hostname] = {}
                    self.set_config_values(host_data[hostname], server,
                                           var_data)
            if isinstance(grp, dict):
                host = self.get_hostname(grp, var_data,
                                         self.DEFAULT_HOSTNAMES)
                hostname_var = host[0]
                hostname = host[1]
                host_data[hostname] = {}
                if '__IP__' not in var_data.keys():
                    var_data['__IP__'] = hostname_var
                    host_data[hostname] = {}
                host_data[hostname] = {}
                self.set_config_values(host_data[hostname], grp, var_data)
        return host_data


    def get_host_ips(self, host_data):
        return host_data.keys()

    def get_inventory(self, topo, layout, config):
        host_data = []
        inven_hosts = []
        for res in topo:
            hd = self.get_host_data(res, config)
            if hd:
                host_data.append(hd)
            inven_hosts = self.get_host_ips(hd)
        # adding sections to respective host groups
        host_groups = self.get_layout_host_groups(layout)
        self.add_sections(host_groups)
        # set children for each host group
        self.set_children(layout)
        # set vars for each host group
        self.set_vars(layout)
        # add ip addresses to each host
        self.add_ips_to_groups(inven_hosts, layout)
        self.add_common_vars(host_groups, layout)
        output = StringIO()
        self.config.write(output)
        return output.getvalue()



import os
import sys

from nose.tools import assert_equal
from nose.tools import assert_dict_equal
from nose.tools import assert_dict_contains_subset
from nose.tools import with_setup

from linchpin import LinchpinAPI
from linchpin.cli import LinchpinCli
from linchpin.cli.context import LinchpinCliContext
from linchpin.tests.mockdata.contextdata import ContextData


def test_cli_create():

    lpctx = LinchpinCliContext()
    lpctx.load_config()
    lpctx.load_global_evars()
    lpc = LinchpinCli(lpctx)

    assert_equal(isinstance(lpc, LinchpinCli), True)


def setup_load_config():

    """
    Perform setup of ContextData() object, and run get_temp_filename()
    """

    global cd
    global provider
    global config_path
    global config_data

    provider = 'dummy'

    cd = ContextData()
    cd.load_config_data()
    cd.create_config()
    config_path = cd.get_temp_filename()
    config_data = cd.cfg_data
    cd.write_config_file(config_path)


def setup_lp_api():

    """
    Perform setup of LinchpinContext, lpc.load_config, and LinchPinAPI
    """

    global lpc
    global lpa

    global target
    global pinfile

    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))

    setup_load_config()

    pinfile = 'PinFile'
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    lpc = LinchpinCliContext()
    lpc.load_config(lpconfig=config_path)
    lpc.load_global_evars()
    lpc.setup_logging()
    lpc.workspace = os.path.realpath(mock_path)

    lpa = LinchpinAPI(lpc)


@with_setup(setup_lp_api)
def test_set_cfg():

    test_dict = {'key': 'value', 'key2': 'value2'}

    for k, v in test_dict.items():
        lpa.set_cfg('test', k, v)

    assert_dict_equal(test_dict, lpa.ctx.cfgs.get('test'))


@with_setup(setup_lp_api)
def test_get_cfg_section():

    lpa.set_cfg('test', 'key', 'value')
    lpa_dict = lpa.get_cfg('test')

    assert_dict_equal(lpa_dict, lpa.ctx.cfgs.get('test'))


@with_setup(setup_lp_api)
def test_get_cfg_item():

    lpa.set_cfg('test', 'key', 'value')
    lpa_value = lpa.get_cfg('test', 'key')

    test_dict = lpa.ctx.cfgs.get('test')
    test_value = test_dict.get('key')

    assert_equal(lpa_value, test_value)


@with_setup(setup_lp_api)
def test_set_evar():

    test_evars = {'ekey': 'evalue'}

    for k, v in test_evars.items():
        lpa.set_evar(k, v)
    lpa_evars = lpa.get_cfg('evars')

    assert_dict_contains_subset(test_evars, lpa_evars)


@with_setup(setup_lp_api)
def test_get_all_evars():

    test_evars = {'ekey1': 'evalue1', 'ekey2': 'evalue2'}
    for k, v in test_evars.items():
        lpa.set_evar(k, v)

    lpa_evars = lpa.get_evar()

    assert_dict_contains_subset(test_evars, lpa_evars)


@with_setup(setup_lp_api)
def test_get_evar_item():

    test_evars = {'ekey2': 'evalue2'}
    for k, v in test_evars.items():
        lpa.set_evar(k, v)

    assert_equal(lpa.get_evar(key='ekey2'), test_evars.get('ekey2'))


def main():
    pass


if __name__ == '__main__':
    sys.exit(main())



from cerberus import Validator
from cerberus import errors


class AnyofValidator(Validator):
    def __init__(self, *args, **kwargs):
        super(AnyofValidator, self).__init__(*args, **kwargs)

    def _validate_anyof(self, definitions, field, value):
        """ {'type': 'list', 'logical': 'anyof'} """

        if 'role' not in definitions[0]['schema'].keys():
            return super(AnyofValidator, self)._validate_anyof(definitions,
                                                               field, value)

        valids = 0
        found_valid_role = False
        _errors = errors.ErrorList()
        valid_roles = []
        for i, definition in enumerate(definitions):
            valid_roles.extend(definition['schema']['role']['allowed'])
            if value['role'] not in definition['schema']['role']['allowed']:
                continue

            found_valid_role = True

            schema = {field: definition.copy()}
            for rule in ('allow_unknown', 'type'):
                if rule not in schema[field] and rule in self.schema[field]:
                    schema[field][rule] = self.schema[field][rule]
            if 'allow_unknown' not in schema[field]:
                schema[field]['allow_unknown'] = self.allow_unknown

            validator = self._get_child_validator(
                schema_crumb=(field, 'role', value['role']), schema=schema,
                allow_unknown=True
            )
            if validator(self.document, update=self.update, normalize=False):
                valids += 1
            else:
                _errors.extend(validator._errors)

        if valids == 0:
            if not found_valid_role:
                role_error = errors.ValidationError(
                    ("resource_definitions", 0, "role"),
                    ("res_defs", "schema", "anyof", -1, "schema", "role"),
                    0x44, "rule", valid_roles, value['role'], ""
                )
                self._error([role_error])
            else:
                self._error(validator._errors)



#!/usr/bin/env python

# flake8: noqa

import os
import sys

def get_ip_arp(text, mac):
    # "mac address='52:54:00:cb:8a:d7'/>"
    for line in text:
        if mac in line:
            return line.split("(")[1].split(")")[0]
    return mac

class FilterModule(object):
    ''' A filter to fix interface's name format '''
    def filters(self):
        return {
            'get_ip_arp': get_ip_arp
        }



from tinydb import TinyDB
from tinydb.storages import JSONStorage
from tinydb.operations import add
from tinydb.operations import set as tinySet
from tinydb.middlewares import CachingMiddleware
from .basedb import BaseDB


def usedb(func):
    def func_wrapper(*args, **kwargs):
        args[0]._opendb()
        x = func(*args, **kwargs)
        args[0]._closedb()
        return x
    return func_wrapper


class TinyRunDB(BaseDB):

    def __init__(self, conn_str):
        self.name = 'TinyRunDB'
        self.conn_str = conn_str
        self.default_table = 'linchpin'


    def _opendb(self):
        self.middleware = CachingMiddleware(JSONStorage)
        self.middleware.WRITE_CACHE_SIZE = 500
        self.db = TinyDB(self.conn_str, storage=self.middleware,
                         default_table=self.default_table)


    def __str__(self):
        if self.conn_str:
            return "{0} at {2}".format(self.name, self.conn_str)
        return "{0} at {1}".format(self.name, 'None')


    @property
    def schema(self):
        return self._schema


    @schema.setter
    def schema(self, schema):
        self._schema = dict()
        self._schema.update(schema)


    @usedb
    def init_table(self, table):
        t = self.db.table(name=table)
        return t.insert(self.schema)


    @usedb
    def update_record(self, table, run_id, key, value):
        t = self.db.table(name=table)
        # get transaction record
        tx_rec = t.get(eid=run_id).get("outputs", [])
        if len(tx_rec) > 0 and isinstance(value, list):
            # fetch the resources dict, index
            # by filtering them from outputs list
            res_list = [(idx, x) for idx, x in enumerate(tx_rec)
                        if "resources" in x]
            if len(res_list) != 0:
                res_idx = res_list[0][0]
                resources = res_list[0][1]
                if "resources" in value[0].keys():
                    de = resources["resources"]
                    for i in value[0]["resources"]:
                        de.append(i)
                    de = {"resources": de}
                    tx_rec[res_idx] = de
                    res = t.update(tinySet(key, [de]), eids=[run_id])
                    return res
        return t.update(add(key, value), eids=[run_id])


    @usedb
    def get_tx_record(self, tx_id):

        t = self.db.table(name='linchpin')
        return t.get(eid=tx_id)


    @usedb
    def get_tx_records(self, tx_ids):

        txs = {}
        t = self.db.table(name='linchpin')
        for tx_id in tx_ids:
            txs[tx_id] = t.get(eid=tx_id)

        return txs


    @usedb
    def get_record(self, table, action='up', run_id=None):

        t = self.db.table(name=table)
        if not run_id:
            run_id = len(t.all())
            if not run_id:
                return (None, 0)

            for rid in range(int(run_id), 0, -1):
                record = t.get(eid=int(rid))
                if record and record['action'] == action:
                    return (record, int(rid))
        else:
            record = t.get(eid=int(run_id))
            if record:
                return(record, int(run_id))


        return (None, 0)


    @usedb
    def get_records(self, table, count=10):
        records = {}
        if table in self.db.tables():
            t = self.db.table(name=table)
            if len(t.all()):
                start = len(t)
                if count == 'all':
                    end = 0
                else:
                    end = start - count
                for i in range(start, end, -1):
                    records[i] = t.get(doc_id=i)
        return records


    @usedb
    def get_tables(self):

        tables = self.db.tables()
        tables.remove(self.default_table)

        return tables


    def remove_record(self, table, key, value):
        pass


    def search(self, table, key=None):
        t = self.db.table(name=table)
        if key:
            return t.search(key)
        return t.all()


    def query(self, table, query):
        pass


    def purge(self, table=None):
        if table:
            return self.db.purge_table(table)
        return self.db.purge_tables()


    def _closedb(self):
        self.db.close()



#!/usr/bin/env python

import os
import json

from linchpin.exceptions import ValidationError
from linchpin.exceptions import ValidationErrorHandler
from linchpin.exceptions import SchemaError
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import TopologyError

from linchpin.validator.anyofvalidator import AnyofValidator


class Validator(object):
    SECTIONS = ['topology', 'layout', 'cfgs', 'hooks']

    def __init__(self, ctx, pb_path, pb_ext):
        self.ctx = ctx
        self.pb_path = pb_path
        self.pb_ext = pb_ext


    def validate(self, target, old_schema=False):
        """
        Attemptes to validate a single target in a PinFile

        Returns the groups returned when the topology is validated

        :param target: the target being validated

        :param old_schema: whether or not to validate using the old schema
                           by default
        """

        for field in target.keys():
            if field not in Validator.SECTIONS:
                raise ValidationError("Section '{0}' not a valid top-level "
                                      "PinFile section. valid sections "
                                      "are '{1}'".format(target,
                                                         Validator.SECTIONS))
        if 'topology' not in target.keys():
            raise ValidationError("Each target must have a topology")

        topo_data = target['topology']
        try:
            resources = self.validate_topology(topo_data)
        except (SchemaError, KeyError):
            # if topology fails, try converting from old to new style
            self._convert_topology(topo_data)
            resources = self.validate_topology(topo_data)
        if 'layout' in target.keys():
            self.validate_layout(target['layout'])
        if 'cfgs' in target.keys():
            self.validate_cfgs(target['cfgs'])

        return resources


    def validate_pretty(self, target, name, old_schema=False):
        results = {}
        return_code = 0

        for field in target.keys():
            if field not in Validator.SECTIONS:
                raise ValidationError("Section '{0}' not a valid top-level "
                                      "PinFile section. valid sections "
                                      "are '{1}'".format(target,
                                                         Validator.SECTIONS))
        if 'topology' not in target.keys():
            results['topology'] = "Each target must have a topology"

        err_prefix = "errors:\n"
        topo_data = target['topology']
        try:
            self.validate_topology(topo_data)
        except (SchemaError, KeyError) as e:
            # if topology fails, try converting from old to new style
            try:
                self._convert_topology(topo_data)
                self.validate_topology(topo_data)
            except SchemaError as s:
                if old_schema:
                    # add a tab to the beginning of each line in the
                    # SchemaError and append to the existing error message
                    error = self._format_error(err_prefix, s)
                else:
                    if type(e) == KeyError:
                        error = "\tfield res_defs['type'] is no longer "\
                                "supported.  Please use 'role' instead"
                        error = self._format_error(err_prefix, e)
                    else:
                        error = self._format_error(err_prefix, e)
                results['topology'] = error
                return_code += 1
            else:
                results['topology'] = "valid under old schema"
        except TopologyError as t:
            error = self._format_error(err_prefix, t)
            results['topology'] = error
            return_code += 1
        else:
            results['topology'] = "valid"

        if 'layout' in target.keys():
            layout_data = target['layout']
            try:
                self.validate_layout(layout_data)
            except SchemaError as e:
                error = self._format_error(err_prefix, e)
                results['layout'] = error + '\n'
            else:
                results['layout'] = "valid"

        if 'cfgs' in target.keys():
            cfgs_data = target['cfgs']
            try:
                self.validate_cfgs(cfgs_data)
            except SchemaError as e:
                error = self._format_error(err_prefix, e)
                results['cfgs'] = error + '\n'
            else:
                results['cfgs'] = "valid"

        return return_code, results


    def _format_error(self, error, e):
        for line in iter(str(e).splitlines(True)):
            error += "\t" + line
        return error


    def validate_topology(self, topo_data):
        """
        Validate the provided topology against the schema

        ;param topo_data: topology dictionary
        """

        # validate high-level topology-components
        self.validate_topology_highlevel(topo_data)

        # validate each resource group
        res_grps = topo_data.get('resource_groups')
        resources = []
        for group in res_grps:
            self.validate_resource_group(group)
            resources.append(group)

        return resources


    def validate_topology_highlevel(self, topo_data):
        """
        validate the higher-level components of the topology

        These are not specific to the provider and must be validated separately
        from the items within each resource group

        :param topo_data topology data from the pinfile
        """

        pb_path = self._find_playbook_path("layout")
        try:
            sp = "{0}/roles/common/files/topo-schema.json".format(pb_path)
            schema = json.load(open(sp))
        except Exception as e:
            raise LinchpinError("Error with schema: '{0}'"
                                " {1}".format(sp, e))

        document = {'topology': topo_data}
        v = AnyofValidator(schema, error_handler=ValidationErrorHandler)

        if not v.validate(document):
            try:
                err = self._gen_error_msg("", "", v.errors)
                raise TopologyError(err)
            except NotImplementedError as e:
                # we shouldn't have this issue using cererus >= 1.2, but
                # this is here just in case an older version has to be used
                self.ctx.log_state("There was an error validating your schema,\
                      but we can't seem to format it for you")
                self.ctx.log_state("Here's the raw error data in case you want\
                      to go through it by hand:")
                self.ctx.log_state(v._errors)


    def validate_resource_group(self, res_grp):
        """
        validate the provided resource group against the schema

        :param res_grp: resource group
        """
        res_grp_type = (res_grp.get('resource_group_type') or
                        res_grp.get('res_group_type'))

        pb_path = self._find_playbook_path(res_grp_type)

        try:
            sp = "{0}/roles/{1}/files/schema.json".format(pb_path,
                                                          res_grp_type)
            schema = json.load(open(sp))
        except Exception as e:
            raise LinchpinError("Error with schema: '{0}'"
                                " {1}".format(sp, e))

        res_defs = res_grp.get('resource_definitions')

        # preload this so it will validate against the schema
        document = {'res_defs': res_defs}
        v = AnyofValidator(schema,
                           error_handler=ValidationErrorHandler)

        if not v.validate(document):
            try:
                err = self._gen_error_msg("", "", v.errors)
                raise SchemaError(err)
            except NotImplementedError as e:
                # we shouldn't have this issue using cererus >= 1.2, but
                # this is here just in case an older version has to be used
                self.ctx.log_state("There was an error validating your schema,\
                      but we can't seem to format it for you")
                self.ctx.log_state("Here's the raw error data in case you want\
                      to go through it by hand:")
                self.ctx.log_state(v._errors)

        return res_grp


    def validate_layout(self, layout_data):
        """
        Validate the provided layout against the schema

        :param layout: layout dictionary
        """

        pb_path = self._find_playbook_path("layout")
        try:
            sp = "{0}/roles/common/files/schema.json".format(pb_path)

            schema = json.load(open(sp))
        except Exception as e:
            raise LinchpinError("Error with schema: '{0}' {1}".format(e))

        v = AnyofValidator(schema)

        if not v.validate(layout_data):
            raise SchemaError('Schema validation failed: {0}'.format(v.errors))


    def validate_cfgs(self, cfgs_data):
        pass


    def _find_playbook_path(self, playbook):
        """
        returns the full path to a given playbook

        :params playbook: name of the playbook
        """

        for path in self.pb_path:
            p = '{0}/{1}{2}'.format(path, playbook, self.pb_ext)

            if os.path.exists(os.path.expanduser(p)):
                return path

        raise LinchpinError("playbook '{0}' not found in"
                            " path: {1}".format(playbook, self.pb_path))


    def _gen_error_msg(self, prefix, section, error):
        """
        Recursively generate a nicely-formatted validation error

        :param prefix:

        :param section: the section in which the error occured

        :param error: the error message itself
        """

        # set the prefix for this subtree
        if section != "":
            if prefix != "":
                prefix += '[' + str(section) + ']'
            else:
                prefix = str(section)

        if isinstance(error, str):
            if prefix == "":
                return error
            return prefix + ": " + error + os.linesep
        elif isinstance(error, list):
            msg = ""
            for i, e in enumerate(error):
                # we don't need to change the prefix here
                msg += self._gen_error_msg(prefix, "", e)
            return msg
        else:
            # in this case, error is a dict
            msg = ""
            for key, val in error.iteritems():
                msg += self._gen_error_msg(prefix, key, val)
            return msg


    def _convert_topology(self, topology):
        """
        For backward compatiblity, convert the old topology format
        into the new format. Should be pretty straightforward and simple.

        ;param topology: topology dictionary
        """
        try:
            res_grps = topology.get('resource_groups')
            if res_grps:
                for res_grp in res_grps:
                    if 'res_group_type' in res_grp.keys():
                        res_grp['resource_group_type'] = (
                            res_grp.pop('res_group_type'))

                    if 'res_defs' in res_grp.keys():
                        res_grp['resource_definitions'] = (
                            res_grp.pop('res_defs'))

                    res_defs = res_grp.get('resource_definitions')
                    if not res_defs:
                        # this means it's either a beaker or openshift topology
                        res_grp_type = res_grp.get('resource_group_type')

                        res_group = self._fix_broken_topologies(res_grp,
                                                                res_grp_type)
                        res_defs = res_group.get('resource_definitions')
                        res_grp['resource_definitions'] = res_defs

                    if res_defs:
                        for res_def in res_defs:
                            if 'res_name' in res_def.keys():
                                res_def['name'] = res_def.pop('res_name')
                            if 'type' in res_def.keys():
                                res_def['role'] = res_def.pop('type')
                            if 'res_type' in res_def.keys():
                                res_def['role'] = res_def.pop('res_type')
                            if 'count' in res_def.keys():
                                res_def['count'] = int(res_def.pop('count'))
                    else:
                        raise TopologyError("'resource_definitions' do not"
                                            " validate in topology"
                                            " ({0})".format(topology))
            else:
                raise TopologyError("'resource_groups' do not validate"
                                    " in topology ({0})".format(topology))

        except Exception:
            raise LinchpinError("Unknown error converting schema. Check"
                                " template data")



#!/usr/bin/env python

import os
import logging

from linchpin.version import __version__
from linchpin.exceptions import LinchpinError

# FIXME: remove this later when not using python2.6
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

try:
    import configparser as ConfigParser
except ImportError:
    import ConfigParser as ConfigParser


class LinchpinContext(object):

    """
    LinchpinContext object, which will be used to manage the cli,
    and load the configuration file.
    """


    def __init__(self):
        """
        Initializes basic variables
        """

        self.version = __version__
        self.verbosity = 1

        self.lib_path = '{0}'.format(os.path.dirname(
                                     os.path.realpath(__file__)))

        self.cfgs = {}
        self._load_constants()


    def _load_constants(self):
        """
        Create self.cfgs with defaults from the linchpin constants file.
        """

        constants_file = '{0}/linchpin.constants'.format(self.lib_path)
        constants_file = os.path.realpath(os.path.expanduser(constants_file))
        self._parse_config(constants_file)


    def load_config(self, search_path=None):
        """
        Update self.cfgs from the linchpin configuration file (linchpin.conf).

        NOTE: Must be implemented by a subclass
        """

        pass


    def _parse_config(self, path):
        """
        Parse configs into the self.cfgs dict from provided path.

        :param path: A path to a config to parse
        """

        try:
            config = ConfigParser.ConfigParser()
            f = open(path)
            config.read_file(f)
            f.close()

            for section in config.sections():
                if not self.cfgs.get(section):
                    self.cfgs[section] = {}

                for k in config.options(section):
                    if section == 'evars':
                        try:
                            self.cfgs[section][k] = (
                                config.getboolean(section, k)
                            )
                        except ValueError:
                            self.cfgs[section][k] = config.get(section, k)
                    else:
                        try:
                            self.cfgs[section][k] = config.get(section, k)
                        except ConfigParser.InterpolationMissingOptionError:
                            value = config.get(section, k, raw=True)
                            self.cfgs[section][k] = value.replace('%%', '%')

        except ConfigParser.InterpolationSyntaxError as e:
            raise LinchpinError('Unable to parse configuration file properly:'
                                ' {0}'.format(e))


    def load_global_evars(self):

        """
        Instantiate the evars variable, then load the variables from the
        'evars' section in linchpin.conf. This will then be passed to
        invoke_linchpin, which passes them to the Ansible playbook as needed.

        """

        self.evars = self.cfgs.get('evars', {})


    def get_cfg(self, section=None, key=None, default=None):
        """
        Get cfgs value(s) by section and/or key, or the whole cfgs object

        :param section: section from ini-style config file

        :param key: key to get from config file, within section

        :param default: default value to return if nothing is found.
        Does not apply if section is not provided.
        """

        if section:
            s = self.cfgs.get(section, default)
            if key and s:
                return self.cfgs[section].get(key, default)
            return s
        return self.cfgs


    def set_cfg(self, section, key, value):
        """
        Set a value in cfgs. Does not persist into a file,
        only during the current execution.

        :param section: section within ini-style config file
        :param key: key to use
        :param value: value to set into section within config file
        """

        if not self.cfgs.get(section):
            self.cfgs.update({section: {}})

        self.cfgs[section][key] = value


    def get_evar(self, key=None, default=None):
        """
        Get the current evars (extra_vars)

        :param key: key to use

        :param default: default value to return if nothing is found
        (default: None)
        """

        if key:
            return self.evars.get(key, default)

        return self.evars


    def set_evar(self, key, value):
        """
        Set a value into evars (extra_vars). Does not persist into a file,
        only during the current execution.

        :param key: key to use

        :param value: value to set into evars
        """

        self.set_cfg('evars', key, value)


    def setup_logging(self):

        """
        Setup logging to the console only

        .. attention:: Please implement this function in a subclass

        """

        self.console = logging.getLogger('lp_console')
        self.console.setLevel(logging.INFO)

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        formatter = logging.Formatter('%(levelname)s %(asctime)s %(message)s')
        ch.setFormatter(formatter)

        self.console.addHandler(ch)


    def log(self, msg, **kwargs):
        """
        Logs a message to a logfile

        :param msg: message to output to log

        :param level: keyword argument defining the log level

        """

        lvl = kwargs.get('level')

        if lvl is None:
            lvl = logging.INFO

        self.console.log(logging.INFO, msg)


    def log_info(self, msg):
        """Logs an INFO message"""

        self.log(msg, level=logging.INFO)


    def log_debug(self, msg):
        """Logs a DEBUG message"""

        self.log(msg, level=logging.DEBUG)


    def log_state(self, msg):
        """
        Logs nothing, just calls pass

        .. attention:: state messages need to be implemented in a subclass
        """

        pass



#!/usr/bin/env python

import os
import ast
import errno
import logging

from linchpin.context import LinchpinContext


class LinchpinCliContext(LinchpinContext):
    """
    Context object, which will be used to manage the cli,
    and load the configuration file
    """


    def __init__(self):
        """
        Initializes basic variables
        """

        # The following values are set in the parent class
        #
        # self.version = __version__
        # self.verbosity = 1
        #
        # self.lib_path = '{0}'.format(os.path.dirname(
        #                              os.path.realpath(__file__)))
        #
        # self.cfgs = {}
        # Load constants from linchpin.constants file.
        # self._load_constants()

        LinchpinContext.__init__(self)


    def load_config(self, lpconfig=None):
        """
        Update self.cfgs from the linchpin configuration file (linchpin.conf).

        The following paths are used to find the config file.
        The search path defaults to the first-found order::

          * /etc/linchpin.conf
          * /linchpin/library/path/linchpin.conf
          * <workspace>/linchpin.conf

        An alternate search_path can be passed.

        :param search_path: A list of paths to search a linchpin config
        (default: None)
        """

        if not self.workspace:
            self.workspace = os.path.realpath(os.path.curdir)

        expanded_path = None

        if lpconfig:
            CONFIG_PATH = [lpconfig]
        else:
            CONFIG_PATH = [
                '/etc/linchpin.conf',
                '~/.config/linchpin/linchpin.conf',
                '{0}/linchpin.conf'.format(self.workspace)
            ]

        existing_paths = []
        for path in CONFIG_PATH:
            expanded_path = (
                "{0}".format(os.path.realpath(os.path.expanduser(path))))

            # implement first found
            if os.path.exists(expanded_path):
                # logging before the config file is setup doesn't work
                # if messages are needed before this, use print.
                existing_paths.append(expanded_path)

        for path in existing_paths:
            self._parse_config(path)


    @property
    def pinfile(self):

        """
        getter function for pinfile name
        """

        return self.get_cfg('init', 'pinfile')


    @pinfile.setter
    def pinfile(self, pinfile):

        """
        setter for workspace
        """

        self.set_cfg('init', 'pinfile', pinfile)


    @property
    def workspace(self):

        """
        getter function for workspace
        """

        return self.get_cfg('lp', 'workspace')


    @workspace.setter
    def workspace(self, workspace):

        """
        setter for workspace
        """

        self.set_cfg('lp', 'workspace', workspace)
        self.set_evar('workspace', workspace)


    def setup_logging(self):

        """
        Setup logging to a file, console, or both. Modifying the `linchpin.conf`
        appropriately will provide functionality.

        """

        self.enable_logging = ast.literal_eval(
            self.cfgs['logger'].get('enable', 'True'))

        if self.enable_logging:

            # create logger
            self.logger = logging.getLogger('lp_logger')
            self.logger.setLevel(eval(self.cfgs['logger'].get('level',
                                                              'logging.DEBUG')))

            logfile = os.path.realpath(os.path.expanduser(
                self.cfgs['logger'].get('file', 'linchpin.log')))

            logdir = os.path.dirname(logfile)

            if not os.path.exists(logdir):
                try:
                    os.makedirs(logdir)
                except OSError as exc:
                    if (exc.errno == errno.EEXIST and
                            os.path.isdir(logdir)):
                        pass
                    else:
                        raise


            fh = logging.FileHandler(logfile)
            fh.setLevel(eval(self.cfgs['logger'].get('level',
                                                     'logging.DEBUG')))
            formatter = logging.Formatter(
                self.cfgs['logger'].get('format',
                                        '%(levelname)s'
                                        ' %(asctime)s %(message)s'))
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)


        self.console = logging.getLogger('lp_console')
        self.console.setLevel(eval(self.cfgs['console'].get('level',
                                                            'logging.INFO')))

        ch = logging.StreamHandler()
        ch.setLevel(eval(self.cfgs['console'].get('level',
                                                  'logging.INFO')))
        formatter = logging.Formatter(
            self.cfgs['console'].get('format', '%(message)s'))
        ch.setFormatter(formatter)
        self.console.addHandler(ch)


    def log(self, msg, **kwargs):
        """
        Logs a message to a logfile or the console

        :param msg: message to log

        :param lvl: keyword argument defining the log level

        :param msg_type: keyword argument giving more flexibility.

        .. note:: Only msg_type `STATE` is currently implemented.
        """

        lvl = kwargs.get('level')
        msg_type = kwargs.get('msg_type')

        if lvl is None:
            lvl = logging.INFO

        if self.verbosity > 1 and not msg_type:
            self.console.log(logging.INFO, msg)

        state_msg = msg
        if msg_type == 'STATE':
            state_msg = 'STATE - {0}'.format(msg)
            self.console.log(logging.INFO, msg)

        if self.enable_logging:
            self.logger.log(lvl, state_msg)


    def log_state(self, msg):
        """Logs a message to stdout"""

        self.log(msg, msg_type='STATE', level=logging.DEBUG)

    def log_info(self, msg):
        """Logs an INFO message """
        self.log(msg, level=logging.INFO)

    def log_debug(self, msg):
        """Logs a DEBUG message"""
        self.log(msg, level=logging.DEBUG)



#!/usr/bin/env python

import os
import ast
import sys
import json
import time
import errno
import hashlib

from uuid import getnode as get_mac
from collections import OrderedDict
from six import text_type

from linchpin.ansible_runner import ansible_runner

from linchpin.hooks.state import State
from linchpin.hooks import LinchpinHooks

from linchpin.rundb.basedb import BaseDB
from linchpin.rundb.drivers import DB_DRIVERS

from linchpin.exceptions import ActionError
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import SchemaError
from linchpin.exceptions import ValidationError

from linchpin.validator import Validator

from linchpin.InventoryFilters import GenericInventory


class LinchpinAPI(object):

    def __init__(self, ctx):
        """
        LinchpinAPI constructor

        :param ctx: context object from context.py

        """

        self.__meta__ = "API"

        self.ctx = ctx

        self.hook_state = None
        self._hook_observers = []
        self.playbook_pre_states = self.get_cfg('playbook_pre_states',
                                                {'up': 'preup',
                                                 'destroy': 'predestroy'})
        self.playbook_post_states = self.get_cfg('playbook_post_states',
                                                 {'up': 'postup',
                                                  'destroy': 'postdestroy',
                                                  'postinv': 'inventory'})
        self.hooks = LinchpinHooks(self)

        self.target_data = {}

        base_path = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
        pkg = self.get_cfg(section='lp', key='pkg', default='linchpin')
        lp_path = '{0}/{1}'.format(base_path, pkg)
        self.pb_ext = self.get_cfg('extensions', 'playbooks', default='.yml')

        # get external_provider_path
        xp_path = self.get_cfg('lp',
                               'external_providers_path',
                               default='').split(':')

        pb_path = '{0}/{1}'.format(lp_path,
                                   self.get_evar('playbooks_folder',
                                                 default='provision'))
        self.pb_path = [pb_path]

        for path in xp_path:
            self.pb_path.append(os.path.expanduser(path))

        self.set_evar('lp_path', lp_path)
        self.set_evar('pb_path', self.pb_path)
        self.set_evar('from_api', True)
        self.workspace = self.get_evar('workspace')


    def setup_rundb(self):
        """
        Configures the run database parameters, sets them into extra_vars
        """

        rundb_conn = self.get_cfg(section='lp',
                                  key='rundb_conn')
        rundb_type = self.get_cfg(section='lp',
                                  key='rundb_type',
                                  default='TinyRunDB')
        rundb_conn_type = self.get_cfg(section='lp',
                                       key='rundb_conn_type',
                                       default='file')
        self.rundb_hash = self.get_cfg(section='lp',
                                       key='rundb_hash',
                                       default='sha256')

        if rundb_conn_type == 'file':

            rundb_conn_f = rundb_conn.replace('::mac::', str(get_mac()))
            rundb_conn_f = rundb_conn_f.replace('{{ workspace }}',
                                                self.workspace)
            rundb_conn_f = os.path.realpath(os.path.expanduser(rundb_conn_f))
            rundb_conn_dir = os.path.dirname(rundb_conn_f)

            if not os.path.exists(rundb_conn_dir):
                try:
                    os.makedirs(rundb_conn_dir)
                except OSError as exc:
                    if (exc.errno == errno.EEXIST and
                            os.path.isdir(rundb_conn_dir)):
                        pass
                    else:
                        raise

        self.set_evar('rundb_type', rundb_type)
        self.set_evar('rundb_conn', rundb_conn_f)
        self.set_evar('rundb_hash', self.rundb_hash)

        return BaseDB(DB_DRIVERS[rundb_type], rundb_conn_f)


    def get_cfg(self, section=None, key=None, default=None):
        """
        Get cfgs value(s) by section and/or key, or the whole cfgs object

        :param section: section from ini-style config file

        :param key: key to get from config file, within section

        :param default: default value to return if nothing is found.
        """

        return self.ctx.get_cfg(section=section, key=key, default=default)


    def set_cfg(self, section, key, value):
        """
        Set a value in cfgs. Does not persist into a file,
        only during the current execution.

        :param section: section within ini-style config file

        :param key: key to use

        :param value: value to set into section within config file
        """

        self.ctx.set_cfg(section, key, value)


    def get_evar(self, key=None, default=None):
        """
        Get the current evars (extra_vars)

        :param key: key to use

        :param default: default value to return if nothing is found
        (default: None)
        """

        return self.ctx.get_evar(key=key, default=default)


    def set_evar(self, key, value):
        """
        Set a value into evars (extra_vars). Does not persist into a file,
        only during the current execution.

        :param key: key to use

        :param value: value to set into evars
        """

        self.ctx.set_evar(key, value)


    @property
    def hook_state(self):
        """
        getter function for hook_state property of the API object
        """

        return self.hook_state


    @hook_state.setter
    def hook_state(self, hook_state):
        """
        hook_state property setter , splits the hook_state string in
        subhook_state and sets linchpin.hook_state object

        :param hook_state: valid hook_state string mentioned in linchpin.conf
        """

        # call run_hooks after hook_state is being set
        if hook_state is None:
            return
        else:
            self.ctx.log_debug('hook {0} initiated'.format(hook_state))
            self._hook_state = State(hook_state, None, self.ctx)

            for callback in self._hook_observers:
                callback(self._hook_state)


    def bind_to_hook_state(self, callback):
        """
        Function used by LinchpinHooksclass to add callbacks

        :param callback: callback function
        """

        self._hook_observers.append(callback)


    def lp_journal(self, view='target', targets=[],
                   fields=None, count=1, tx_ids=None):

        rundb = self.setup_rundb()

        journal = {}

        if not len(targets):
            targets = rundb.get_tables()

        if view == 'target':
            # initialize rundb table
            dateformat = self.get_cfg('logger',
                                      'dateformat',
                                      default='%m/%d/%Y %I:%M:%S %p')
            for target in targets:
                tgt_records = rundb.get_records(table=target, count=count)

                if fields and 'start' in fields:
                    for run_id, record in tgt_records.iteritems():
                        st = record.get('start')
                        strt = float(st) if len(st) else float(1000000000.0)
                        start = time.localtime(strt)
                        record['start'] = time.strftime(dateformat, start)

                if fields and 'end' in fields:
                    for run_id, record in tgt_records.iteritems():
                        en = record.get('end')
                        endt = float(st) if len(en) else float(1000000000.0)
                        end = time.localtime(endt)
                        record['end'] = time.strftime(dateformat, end)


                journal[target] = tgt_records
        if view == 'tx':
            if len(tx_ids):
                journal = rundb.get_tx_records(tx_ids)
            else:
                journal = rundb.get_records('linchpin', count=count)

        return journal


    def _find_playbook_path(self, playbook):

        for path in self.pb_path:
            p = '{0}/{1}{2}'.format(path, playbook, self.pb_ext)

            if os.path.exists(os.path.expanduser(p)):
                return path

        raise LinchpinError("playbook '{0}' not found in"
                            " path: {1}".format(playbook, self.pb_path))


    def _fix_broken_topologies(self, res_grp, res_grp_type):
        """
        Because the new tooling requires resource_definitions, both beaker
        and openshift didn't previously suport this section in topologies.
        This function is called if resource_definitions are not included
        in the topology. The topology will be given a resource_definitions
        section.

        :param res_grp: The resource group to update

        :param res_grp_type: Which type to convert (this may not matter)
        """

        res_defs = {}
        # with beaker, there will only be one
        # resource_definition upon conversion
        if res_grp_type == 'beaker':
            res_defs['role'] = 'bkr_server'
            res_defs['whiteboard'] = res_grp.pop('whiteboard',
                                                 'Provisioned with LinchPin')
            res_defs['job_group'] = res_grp.pop('job_group')
            res_defs['recipesets'] = res_grp.pop('recipesets')
            res_defs['cancel_message'] = res_grp.pop('cancel_message',
                                                     'Canceled by LinchPin')
            res_defs['max_attempts'] = res_grp.pop('max_attempts', 60)
            res_defs['attempt_wait_time'] = res_grp.pop('attempt_wait_time',
                                                        60)
            res_grp['resource_definitions'] = [res_defs]

            return res_grp

        if res_grp_type == 'openshift':
            res_defs = res_grp.pop('resources')

            count = 0
            for res in res_defs:
                res['name'] = 'openshift-res_{0}'.format(str(count))
                count += 1
                if res.get('inline_data'):
                    res['role'] = 'openshift_inline'
                    res['data'] = res.pop('inline_data')
                if res.get('file_reference'):
                    res['role'] = 'openshift_external'
                    res['filiename'] = res.pop('file_reference')

            res_grp['resource_definitions'] = res_defs

            # the old openshift role put credentials in the
            # resource group definition. Moving to credentials
            creds = {}
            creds['api_endpoint'] = res_grp.pop('api_endpoint')
            creds['api_token'] = res_grp.pop('api_token')
            res_grp['credentials'] = creds

            return res_grp


    def _convert_layout(self, layout_data):
        """
        Convert the layout to retain order of the layout hosts
        ;param layout_data: layout_data
        """
        layout_json = layout_data
        layout_hosts = []
        ihosts = layout_data["inventory_layout"]["hosts"]
        for k in ihosts:
            layout_host = {}
            layout_host["name"] = k
            for key in ihosts[k]:
                layout_host[key] = ihosts[k][key]
            layout_hosts.append(layout_host)
        layout_json["inventory_layout"]["hosts"] = layout_hosts
        return layout_json


    def generate_inventory(self, resource_data, layout, inv_format="cfg",
                           topology_data={}, config_data={}):
        inv = GenericInventory.GenericInventory(inv_format=inv_format)
        inventory = inv.get_inventory(resource_data, layout, topology_data,
                                      config_data)
        return inventory

    def get_pf_data_from_rundb(self, targets, run_id=None, tx_id=None):
        """
        This function takes the action and provision_data, returns the
        pinfile data

        :param targets: A list of targets for which to get the data

        :param targets: Tuple of target(s) for which to gather data.

        :param run_id: run_id associated with target (Default: None)

        :param tx_id: tx_id for which to gather data (Default: None)
        """

        rundb = self.setup_rundb()

        if run_id and tx_id:
            raise ActionError("'run_id' and 'tx_id' are mutually exclusive")

        pf_data = {}
        pinfile = OrderedDict()

        if run_id:
            for target in targets:
                pf_data[target] = rundb.get_record(target,
                                                   action='up',
                                                   run_id=run_id)
        if tx_id:
            record = rundb.get_tx_record(tx_id)

            if not record or not len(record):
                return None

            if len(targets):
                for tgts in record['targets']:
                        for tgt, data in tgts.iteritems():
                            run_id = int(data.keys()[0])
                            if tgt in targets:
                                tgt_data = (rundb.get_record(tgt,
                                            action=record['action'],
                                            run_id=run_id))
                                pf_data[tgt] = tgt_data
            else:
                for tgts in record['targets']:
                    for tgt, data in tgts.iteritems():
                        run_id = int(data.keys()[0])
                        tgt_data = (rundb.get_record(tgt,
                                    action=record['action'],
                                    run_id=run_id))
                        pf_data[tgt] = tgt_data


        for t, data in pf_data.iteritems():
            topo_data = data[0]['inputs'][0].get('topology_data')
            layout_data = data[0]['inputs'][0].get('layout_data')
            hooks_data = data[0]['inputs'][0].get('hooks_data')

            pinfile[t] = {}
            pinfile[t]['topology'] = topo_data
            pinfile[t]['run_id'] = data[1]
            if layout_data:
                pinfile[t]['layout'] = layout_data
            if hooks_data:
                pinfile[t]['hooks'] = hooks_data

        return pinfile


    def do_action(self, provision_data, action='up', run_id=None, tx_id=None):
        """
        This function takes provision_data, and executes the given
        action for each target within the provision_data disctionary.

        :param provision_data: PinFile data as a dictionary, with
        target information

        :param action: Action taken (up, destroy, etc). (Default: up)

        :param run_id: Provided run_id to duplicate/destroy (Default: None)

        :param tx_id: Provided tx_id to duplicate/destroy (Default: None)

        .. .note:: The `run_id` value differs from the `rundb_id`, in that
                   the `run_id` is an existing value in the database.
                   The `rundb_id` value is created to store the new record.
                   If the `run_id` is passed, it is used to collect an existing
                   `uhash` value from the given `run_id`, which is in turn used
                   to perform an idempotent reprovision, or destroy provisioned
                   resources.
        """
        ansible_console = False
        if self.ctx.cfgs.get('ansible'):
            ansible_console = (
                ast.literal_eval(self.get_cfg('ansible',
                                              'console',
                                              default='False')))

        if not ansible_console:
            ansible_console = bool(self.ctx.verbosity)

        results = {}


        return_code = 99
        vault_pass = self.get_evar('vault_pass')
        if vault_pass == '':
            self.set_evar('vault_pass',
                          self.ctx.get_cfg('evars',
                                           'vault_pass',
                                           default=''))

        for target in provision_data.keys():
            if not isinstance(provision_data[target], dict):
                raise LinchpinError("Target '{0}' does not"
                                    " exist.".format(target))

        targets = [x.lower() for x in provision_data.keys()]
        if 'linchpin' in targets:
            raise LinchpinError("Target 'linchpin' is not allowed.")

        for target in provision_data.keys():
            if target == 'cfgs':
                continue

            self.ctx.log_debug("Processing target: {0}".format(target))

            results[target] = {}
            self.set_evar('target', target)

            rundb = self.setup_rundb()

            if tx_id:
                record = rundb.get_tx_record(tx_id)
                run_id = (record['targets'][0][target].keys()[0])


            rundb_schema = json.loads(self.get_cfg(section='lp',
                                      key='rundb_schema'))
            rundb.schema = rundb_schema
            self.set_evar('rundb_schema', rundb_schema)

            start = time.time()
            st_uhash = int(start * 1000)
            uhash = None

            # generate a new rundb_id
            # (don't confuse it with an already existing run_id)
            rundb_id = rundb.init_table(target)
            orig_run_id = rundb_id

            uhash_length = self.get_cfg('lp', 'rundb_uhash_length')
            uhash_len = int(uhash_length)
            if not run_id:
                hash_str = ':'.join([target, str(tx_id), str(rundb_id),
                                     str(st_uhash)])
                if isinstance(hash_str, text_type):
                    hash_str = hash_str.encode('utf-8')
                uh = hashlib.new(self.rundb_hash, hash_str)
                uhash = uh.hexdigest()[:uhash_len]

            if action == 'destroy' or run_id:
                # look for the action='up' records to destroy
                data, orig_run_id = rundb.get_record(target,
                                                     action='up',
                                                     run_id=run_id)

                if data:
                    uhash = data.get('uhash')
                    self.ctx.log_debug("using data from"
                                       " run_id: {0}".format(run_id))

            elif action not in ['up', 'destroy']:
                # it doesn't appear this code will will execute,
                # but if it does...
                raise LinchpinError("Attempting '{0}' action on"
                                    " target: '{1}' failed. Not an"
                                    " action.".format(action, target))


            self.ctx.log_debug('rundb_id: {0}'.format(rundb_id))
            self.ctx.log_debug('uhash: {0}'.format(uhash))

            rundb.update_record(target, rundb_id, 'uhash', uhash)
            rundb.update_record(target, rundb_id, 'start', str(start))
            rundb.update_record(target, rundb_id, 'action', action)

            self.set_evar('orig_run_id', orig_run_id)
            self.set_evar('rundb_id', rundb_id)
            self.set_evar('uhash', uhash)

            try:
                validator = Validator(self.ctx, self.pb_path, self.pb_ext)
                resources = validator.validate(provision_data[target])
            except SchemaError:
                raise ValidationError("Target '{0}' does not validate.  For"
                                      " more information run `linchpin "
                                      "validate`".format(target))

            topology_data = provision_data[target].get('topology')
            self.set_evar('topo_data', topology_data)

            rundb.update_record(target,
                                rundb_id,
                                'inputs',
                                [
                                    {'topology_data':
                                     provision_data[target]['topology']}
                                ])

            if provision_data[target].get('layout', None):
                l_data = provision_data[target]['layout']
                provision_data[target]['layout'] = self._convert_layout(l_data)
                self.set_evar('layout_data', provision_data[target]['layout'])

                rundb.update_record(target,
                                    rundb_id,
                                    'inputs',
                                    [
                                        {'layout_data':
                                         provision_data[target]['layout']}
                                    ])

            if provision_data[target].get('hooks', None):
                hooks_data = provision_data[target].get('hooks')
                self.set_evar('hooks_data', hooks_data)
                rundb.update_record(target,
                                    rundb_id,
                                    'inputs',
                                    [
                                        {'hooks_data':
                                         provision_data[target]['hooks']}
                                    ])

            if provision_data.get('cfgs', None):
                vars_data = provision_data[target].get('cfgs')
                self.set_evar('cfgs_data', vars_data)
                rundb.update_record(target,
                                    rundb_id,
                                    'cfgs',
                                    [
                                        {'user':
                                         provision_data['cfgs']}
                                    ])

            # note : changing the state triggers the hooks
            self.hooks.rundb = (rundb, rundb_id)
            self.pb_hooks = self.get_cfg('hookstates', action)
            self.ctx.log_debug('calling: {0}{1}'.format('pre', action))

            if (('pre' in self.pb_hooks) and not self.get_cfg('hook_flags',
                                                              'no_hooks')):
                self.hook_state = '{0}{1}'.format('pre', action)

            # FIXME need to add rundb data for hooks results

            # invoke the appropriate action
            return_code, results[target]['task_results'] = (
                self._invoke_playbooks(resources, action=action,
                                       console=ansible_console)
            )

            # logs the return_code back to console and quits
            # quits if return_code > 0
            self.ctx.log("Playbook Return code {0}".format(str(return_code)))
            if return_code > 0 and action == "up":
                sys.exit(return_code)

            if not return_code:
                self.ctx.log_state("Action '{0}' on Target '{1}' is "
                                   "complete".format(action, target))

            # FIXME Check the result[target] value here, and fail if applicable.
            # It's possible that a flag might allow more targets to run, then
            # return an error code at the end.

            # add post provision hook for inventory generation
            if (('inv' in self.pb_hooks) and not self.get_cfg('hook_flags',
                                                              'no_hooks')):
                self.hook_state = 'postinv'

            if (('post' in self.pb_hooks) and
                    (self.__meta__ == "API") and
                    not self.get_cfg('hook_flags', 'no_hooks')):
                self.hook_state = '{0}{1}'.format('post', action)

            end = time.time()

            rundb.update_record(target, rundb_id, 'end', str(end))
            rundb.update_record(target, rundb_id, 'rc', return_code)

            run_data = rundb.get_record(target,
                                        action=action,
                                        run_id=rundb_id)

            results[target]['rundb_data'] = {rundb_id: run_data[0]}


        # generate the linchpin_id and structure
        lp_schema = ('{"action": "", "targets": []}')

        rundb = self.setup_rundb()
        rundb.schema = json.loads(lp_schema)
        lp_id = rundb.init_table('linchpin')

        summary = {}

        for target, data in results.iteritems():
            for k, v in data['rundb_data'].iteritems():
                summary[target] = {k: {'rc': v['rc'], 'uhash': v['uhash']}}

        rundb.update_record('linchpin', lp_id, 'action', action)
        rundb.update_record('linchpin', lp_id, 'targets', [summary])

        lp_data = {lp_id: {'action': action,
                           'summary_data': summary,
                           'results_data': results}}

        return (return_code, lp_data)


    def do_validation(self, provision_data, old_schema=False):
        """
        This function takes provision_data, and attempts to validate the
        topologies for that data

        :param provision_data: PinFile data as a dictionary, with
        target information
        """

        results = {}


        return_code = 0

        for target in provision_data.keys():
            if not isinstance(provision_data[target], dict):
                raise LinchpinError("Target '{0}' does not"
                                    " exist.".format(target))

        targets = [x.lower() for x in provision_data.keys()]
        if 'linchpin' in targets:
            raise LinchpinError("Target 'linchpin' is not allowed.")

        for target in provision_data.keys():
            self.ctx.log_debug("Processing target: {0}".format(target))

            results[target] = {}
            self.set_evar('target', target)

            validator = Validator(self.ctx, self.pb_path, self.pb_ext)
            target_dict = provision_data[target]
            ret, results[target] = validator.validate_pretty(target_dict,
                                                             target,
                                                             old_schema)
            return_code += ret

        return return_code, results


    def get_run_data(self, tx_id, fields, targets=()):
        """
        Returns the RunDB for data from a specified field given a tx_id.
        The fields consist of the major sections in the RunDB (target
        view only). Those fields are action, start, end, inputs, outputs,
        uhash, and rc.

        :param tx_id: tx_id to search
        :param fields: Tuple of fields to retrieve for each record requested.
        :param targets: Tuple of targets to search from within the tx_ids
        """

        rundb = self.setup_rundb()

        tgt_run_ids = {}
        target_data = {}

        record = rundb.get_tx_record(tx_id)

        if not record or not len(record):
            return None


        # get run_ids to query
        if len(targets):
            for tgts in record['targets']:
                    for tgt, data in tgts.iteritems():
                        if tgt in targets:
                            tgt_run_ids[tgt] = int(data.keys()[0])
        else:
            for tgts in record['targets']:
                for tgt, data in tgts.iteritems():
                    tgt_run_ids[tgt] = int(data.keys()[0])

        for target, run_id in tgt_run_ids.iteritems():
            record = rundb.get_record(target, run_id=run_id, action='up')
            field_data = {}
            single_value_fields = ('action',
                                   'start',
                                   'end',
                                   'rc',
                                   'uhash')

            for field in fields:
                f = record[0].get(field)
                if f:
                    if field in single_value_fields:
                        field_data[field] = f
                    else:
                        data_array = {}
                        for fld in f:
                            for k, v in fld.iteritems():
                                if field == 'outputs':
                                    if isinstance(v, dict):
                                        values = v
                                    else:
                                        values = []
                                        for value in v:
                                            values.append(value)
                                    data_array[k] = values
                                else:
                                    data_array[k] = v

                            field_data[field] = data_array

            target_data[target] = field_data

        return target_data

    def _get_run_data_by_txid(self, tx_id=None):
        rundb = self.setup_rundb()
        latest_run_data = {}
        run_data = {}
        if tx_id is None:
            latest_run_data = rundb.get_records('linchpin', count=1)
            run_data = self.get_run_data(latest_run_data.keys()[0],
                                         ('outputs',
                                          'inputs',
                                          'cfgs'))
        else:
            latest_run_data = rundb.get_records('linchpin',
                                                count='all')
            latest_run_data = {tx_id: latest_run_data.get(tx_id)}
            run_data = self.get_run_data(tx_id,
                                         ('outputs',
                                          'inputs',
                                          'cfgs'))
        for k in latest_run_data:
            v = latest_run_data[k]
            target_group = v.get("targets", [])
            # Note:
            # target_group always returns a dict inside a list due to
            # rundb implemenatation. This code works for one PinFile run
            # might be subjected to change in future releases
            if len(target_group) == 0:
                continue
            else:
                target_group = target_group[0]
            for key in target_group.keys():
                target_group[key].update(run_data.get(key, {}))
            latest_run_data[k]["targets"] = [target_group]
        return latest_run_data

    def _get_module_path(self):
        module_paths = []
        module_folder = self.get_cfg('lp',
                                     'module_folder',
                                     default='library')
        for path in reversed(self.pb_path):
            module_paths.append('{0}/{1}/'.format(path, module_folder))
        return module_paths

    def _find_n_run_pb(self, pb_name, inv_src, console=True):
        pb_path = self._find_playbook_path(pb_name)
        playbook_path = '{0}/{1}{2}'.format(pb_path, pb_name, self.pb_ext)
        extra_vars = self.get_evar()
        return_code, res = ansible_runner(playbook_path,
                                          self._get_module_path(),
                                          extra_vars,
                                          inventory_src=inv_src,
                                          verbosity=self.ctx.verbosity,
                                          console=console)
        return return_code, res

    def _invoke_playbooks(self, resources={}, action='up', console=True,
                          providers=[]):
        """
        Uses the Ansible API code to invoke the specified linchpin playbook

        :param resources: dict of resources to provision
        :param action: Which ansible action to run (default: 'up')
        :param console: Whether to display the ansible console (default: True)
        """

        return_code = 0
        results = []

        self.set_evar('_action', action)
        self.set_evar('state', 'present')

        if action == 'setup' or action == 'ask_sudo_setup':
            self.set_evar('setup_providers', providers)
            return_code, res = self._find_n_run_pb(action,
                                                   "localhost",
                                                   console=console)
            if res:
                results.append(res)
            if not len(results):
                results = None
            return (return_code, results)

        if action == 'destroy':
            self.set_evar('state', 'absent')

        inventory_src = '{0}/localhost'.format(self.workspace)

        for resource in resources:
            self.set_evar('resources', resource)
            playbook = resource.get('resource_group_type')
            return_code, res = self._find_n_run_pb(playbook,
                                                   inventory_src,
                                                   console=console)
            if action == "up" and return_code > 0:
                raise LinchpinError("Unsuccessful provision of resource "
                                    "System return: {0}".format(return_code))
                sys.exit(return_code)

            if res:
                results.append(res)

        if not len(results):
            results = None

        return (return_code, results)



#!/usr/bin/env python

import os
import ast
from setuptools import setup, find_packages

with open('linchpin/version.py') as f:
    for line in f:
        if line.startswith('__version__'):
            ver = ast.parse(line).body[0].value.s
            break

# reading requirements from requirements.txt
dir_path = os.path.dirname(os.path.realpath(__file__))
reqs_file = 'requirements.txt'.format(dir_path)
with open(reqs_file) as f:
    required = f.read().splitlines()

setup_required = list(required)
setup_required.append('pytest-runner')

ignore_dir = ['.git']

setup(
    name='linchpin',
    version=ver,
    description='Ansible based multi cloud provisioner',
    author='samvaran kashyap rallabandi',
    author_email='linchpin@redhat.com',
    url='http://linchpin.readthedocs.io/',
    setup_requires=setup_required,
    install_requires=required,
    entry_points='''
        [console_scripts]
        linchpin=linchpin.shell:runcli
    ''',
    tests_require=["pytest", "nose", "mock", "coverage", "flake8"],
    extras_require={
        'krbV': ["python-krbV"],
        'beaker': ['beaker-client>=23.3', 'python-krbV'],
        'docs': ["docutils", "sphinx", "sphinx_rtd_theme"],
        'tests': ["nose", "mock", "coverage", "flake8"],
        'libvirt': ["libvirt-python>=3.0.0", "lxml"],
    },
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    scripts=[
        'scripts/install_libvirt_deps.sh',
        'scripts/install_selinux_venv.sh'
    ]
)



#!/usr/bin/env python

import os
import ast
import sys
import json
import click
import hashlib

from random import randint
from distutils import dir_util
from collections import OrderedDict
from jinja2 import Environment

from linchpin import LinchpinAPI
from linchpin.fetch import FETCH_CLASS
from linchpin.exceptions import ActionError
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import TopologyError
from linchpin.utils.dataparser import DataParser



class LinchpinCli(LinchpinAPI):

    def __init__(self, ctx):
        """
        Set some variables, pass to parent class
        """

        LinchpinAPI.__init__(self, ctx)
        self.__meta__ = "CLI"
        self.parser = DataParser()


    @property
    def pinfile(self):
        """
        getter function for pinfile name
        """

        return self.ctx.pinfile


    @pinfile.setter
    def pinfile(self, pinfile):
        """
        setter for pinfile name
        """

        self.ctx.pinfile = pinfile


    @property
    def pf_data(self):
        """
        getter for pinfile template data
        """

        return self.ctx.pf_data


    @pf_data.setter
    def pf_data(self, pf_data):
        """
        setter for pinfile template data
        """

        self.ctx.pf_data = pf_data


    @property
    def workspace(self):
        """
        getter function for context workspace
        """

        return self.ctx.workspace


    @workspace.setter
    def workspace(self, workspace):
        """
        setter for context workspace
        """

        self.ctx.workspace = workspace
        self.ctx.set_evar('workspace', workspace)


    def lp_init(self, providers=['libvirt']):
        """
        Initializes a linchpin project. Creates the necessary directory
        structure, includes PinFile, topologies and layouts for the given
        provider. (Default: Dummy. Other providers not yet implemented.)

        :param providers: A list of providers for which templates
        (and a target) will be provided into the workspace.
        NOT YET IMPLEMENTED
        """

        src = self.get_cfg('init', 'source', 'templates/')
        src_w_path = os.path.realpath('{0}/{1}'.format(self.ctx.lib_path, src))

        pf_w_path = self._get_pinfile_path(exists=False)

        # appending .lp_example so we know which PinFile is which
        src_pf = os.path.realpath('{0}.lp_example'.format(pf_w_path))

        try:
            if os.path.exists(pf_w_path):
                if not click.confirm('{0} already exists,'
                                     'overwrite it?'.format(pf_w_path),
                                     default=False):
                    sys.exit(0)

            dir_util.copy_tree(src_w_path, self.workspace, verbose=1)
            os.rename(src_pf, pf_w_path)

            self.ctx.log_state('{0} and file structure created at {1}'.format(
                self.pinfile, self.workspace))
        except Exception as e:
            self.ctx.log_state('Error: {0}'.format(e))
            sys.exit(1)

    def _write_to_inventory(self, tx_id=None, inv_path=None, inv_format="cfg"):
        latest_run_data = self._get_run_data_by_txid()
        if tx_id:
            latest_run_data = self._get_run_data_by_txid(tx_id)
        all_inventories = {}
        try:
            for t_id in latest_run_data:
                targets = latest_run_data[t_id]["targets"][0]
                # if there are multiple targets mentioned in pinfile
                # the multiple inventory files are being generated
                inv_file_count = 0 if len(targets) > 1 else False
                for name in targets:
                    if "layout_data" in targets[name]["inputs"]:
                        lt_data = targets[name]["inputs"]["layout_data"]
                        t_data = targets[name]["inputs"]["topology_data"]
                        c_data = {}
                        if "cfgs" in targets[name].keys():
                            c_data = targets[name]["cfgs"]["user"]
                        i_path = targets[name]["outputs"]["inventory_path"][0]
                        layout = lt_data["inventory_layout"]
                        # check whether inventory_file is mentioned in layout
                        if layout.get("inventory_file", None):
                            i_path = layout["inventory_file"]
                            # if its not absolute path make path
                            # relative to workspace
                            if not os.path.isabs(i_path):
                                # get default variable for inventories_path
                                d_p = self.ctx.get_evar("default_inventories"
                                                        "_path")
                                d_p = Environment().from_string(d_p).render(
                                    workspace=self.ctx.get_cfg('evars',
                                                               'workspace'))
                                i_path = "{0}/{1}".format(d_p, i_path)
                        else:
                            i_path = targets[name]["outputs"]
                            i_path = i_path["inventory_path"][0]

                        if not os.path.exists(os.path.dirname(i_path)):
                            os.makedirs(os.path.dirname(i_path))
                        if inv_path and inv_file_count is not False:
                            i_path = inv_path + str(inv_file_count)
                        # r_o -> resources_outputs
                        r_o = targets[name]["outputs"]["resources"]
                        inv = self.generate_inventory(r_o,
                                                      layout,
                                                      inv_format=inv_format,
                                                      topology_data=t_data,
                                                      config_data=c_data)
                        # if inv_path is explicitly mentioned it is used
                        if inv_path:
                            i_path = inv_path
                        # if there are multiple targets based on
                        # number of targets multiple files are
                        # generated with suffixes
                        if inv_path and isinstance(inv_file_count, int):
                            i_path = inv_path + "." + str(name)
                            with open(i_path, 'w') as the_file:
                                the_file.write(inv)
                            inv_file_count += 1
                        else:
                            with open(i_path, 'w') as the_file:
                                the_file.write(inv)
                        all_inventories[name] = inv
            return all_inventories

        except Exception as e:
            self.ctx.log_state('Error: {0}'.format(e))
            sys.exit(1)
        return True


    def _get_pinfile_path(self, exists=True):
        """
        This function finds the self.pinfile. If the file is a full path,
        it is expanded and used. If not found, the lp.default_pinfile
        configuration value is used for the pinfile, the workspace is
        prepended and returned.

        :param exists:
            Whether the pinfile is supposed to already exist (default: True)
        """

        if not self.pinfile:
            self.pinfile = self.get_cfg('lp',
                                        'default_pinfile',
                                        default='PinFile')

        pf_w_path = os.path.realpath(os.path.expanduser(self.pinfile))
        if self.workspace:
            pf_w_path = '{0}/{1}'.format(self.workspace, self.pinfile)

        # Ensure a PinFile path will exist
        if not os.path.exists(pf_w_path) and exists:
            pf_w_path = '{0}/{1}'.format(self.workspace, self.pinfile)

        # If the PinFile doesn't exist, raise an error
        if not os.path.exists(pf_w_path) and exists:
            raise LinchpinError('{0} not found. Please check that it'
                                ' exists and try again'.format(pf_w_path))

        return pf_w_path


    def _get_data_path(self):
        """
        This function finds the template data path, or returns the data.
        If the file is a full path, it is expanded and used.
        If not found, the workspace is prepended. If still not found, it
        is assumed what is passed in is data and returned to the caller.

        :param data_path:
            Consists of either a absolute path, relative path, or the actual
            template data.
        """

        if not self.pf_data or not self.pf_data.startswith('@'):
            return None
        else:
            pf_data_path = self.pf_data[1:]
            data_w_path = os.path.abspath(os.path.expanduser(pf_data_path))

            if not os.path.exists(data_w_path):
                data_w_path = '{0}/{1}'.format(self.workspace, pf_data_path)
                if not os.path.exists(data_w_path):
                    error_txt = "Template-data (-d) file was not found. Check"
                    error_txt += " the template path and try again."
                    raise TopologyError(error_txt)

            return data_w_path


    def _write_distilled_context(self, run_data):
        """
        This method takes all of the provided run_data, loops through the
        distiller section of the linchpin.constants and writes out the
        linchpin.context (name TBD) to the 'resources' directory.
        """

        dist_roles = self.get_cfg('distiller')

        resources_folder = self.get_evar('resources_folder',
                                         default='resources')
        context_path = '{0}/{1}'.format(self.workspace, resources_folder)

        resources_path = os.path.expanduser(
            self.get_evar('default_resources_path'))

        if (os.path.isabs(resources_path)):
                context_path = resources_path
        if not os.path.exists(context_path):
            os.makedirs(context_path)

        context_file = '{0}/{1}'.format(context_path, 'linchpin.distilled')

        roles = []
        dist_data = {}
        # get roles used
        for target, data in run_data.iteritems():
            inputs = data.get('inputs')
            topo = inputs.get('topology_data')
            res_grps = topo.get('resource_groups')

            for group in res_grps:
                res_defs = group.get('resource_definitions')
                for rd in res_defs:
                    roles.append(rd.get('role'))

            fields = {}
            outputs = data.get('outputs', {})
            resources = outputs.get('resources', [])
            for dist_role, flds in dist_roles.iteritems():
                if dist_role in roles:
                    for f in flds.split(','):
                        if '.' not in f:
                            fld = fields.get('single', [])
                            fields[f] = None
                        else:
                            k, v = f.split('.')
                            fld = fields.get(k, [])
                            if ':' in v:
                                subfields = {}
                                k2, val = v.split(':')
                                subfld = subfields.get(k2, [])
                                subfld.append(val)
                                subfields[k2] = subfld
                                fld.append(subfields)
                                fields[k] = fld
                            else:
                                fld.append(v)
                                fields[k] = fld

            try:
                for res in resources:
                    res_data = []
                    for k, v in fields.iteritems():
                        if not v:
                            res_dict = {}
                            res_dict[k] = res.get(k)
                            res_data.append(res_dict)
                        else:
                            for rsrc in res.get(k, []):
                                res_dict = {}
                                for value in v:
                                    if isinstance(value, dict):
                                        for key, vals in value.iteritems():
                                            subrsc = rsrc.get(key)
                                            for val in vals:
                                                res_dict[val] = subrsc.get(val)
                                    else:
                                        res_dict[value] = rsrc.get(value)
                                res_data.append(res_dict)

                    if target not in dist_data.keys():
                        dist_data[target] = []

                    if len(res_data) and res_data not in dist_data[target]:
                        dist_data[target].extend(res_data)
            except Exception as e:
                self.ctx.log_info('Error recording distilled context'
                                  ' ({0})'.format(e))
        with open(context_file, 'w+') as f:
            f.write(json.dumps(dist_data))

    def _write_latest_run(self):
        latest_run_data = self._get_run_data_by_txid()
        resources_path = self.get_evar('resources_folder')
        context_path = '{0}/{1}'.format(self.workspace, resources_path)
        if not os.path.exists(context_path):
            os.makedirs(context_path)
        context_file = '{0}/{1}'.format(context_path, 'linchpin.latest')
        with open(context_file, 'w+') as f:
            f.write(json.dumps(latest_run_data))


    def lp_down(self, pinfile, targets=(), run_id=None):
        """
        This function takes a list of targets, and performs a shutdown on
        nodes in the target's topology. Only providers which support shutdown
        from their API (Ansible) will support this option.

        CURRENTLY UNIMPLEMENTED

        .. seealso:: lp_destroy

        :param pinfile:
            Provided PinFile, with available targets,

        :param targets:
            A tuple of targets to provision.
        """

        pass


    def lp_up(self, targets=(), run_id=None, tx_id=None, inv_f="cfg"):
        """
        This function takes a list of targets, and provisions them according
        to their topology.

        :param targets:
            A tuple of targets to provision

        :param run_id:
            An optional run_id if the task is idempotent

        :param tx_id:
            An optional tx_id if the task is idempotent
        """

        # Prep input data

        # Execute prepped data
        return_code, return_data = self._execute_action('up',
                                                        targets,
                                                        run_id=run_id,
                                                        tx_id=tx_id)

        # Distill data
        new_tx_id = return_data.keys()[0]

        # This is what the API allows.
        # run_data = self.get_run_data(new_tx_id, ('outputs', 'inputs',
        #                                          'action', 'cfgs', 'start',
        #                                          'end', 'rc', 'uhash'))

        run_data = self.get_run_data(new_tx_id, ('inputs', 'outputs'))

        # Export distilled data in useful ways
        # # Write out run_data to a file for now
        distill_data = self.get_cfg('lp', 'distill_data')
        gen_resources = self.get_evar('generate_resources')

        if ast.literal_eval(distill_data.title()) and not gen_resources:
            if return_code:
                distill_on_error = self.get_cfg('lp',
                                                'distill_on_error',
                                                default='False')
                if ast.literal_eval(distill_on_error.title()):
                    self._write_distilled_context(run_data)
            else:
                    self._write_distilled_context(run_data)
        self._write_latest_run()
        self._write_to_inventory(inv_format=inv_f)

        if (('post' in self.pb_hooks) and
                (self.__meta__ == "CLI") and
                not self.get_cfg('hook_flags', 'no_hooks')):
            self.hook_state = '{0}{1}'.format('post', 'up')

        # Show success and errors, with data
        return (return_code, return_data)


    def lp_validate(self, targets=(), old_schema=False):
        """
        This function takes a list of targets, and validates their topology.

        :param targets:
            A tuple of targets to provision

        :param old_schema
            Denotes whether schema should be validated with the old schema
            rather than the new one!/usr/bin/env python
        """

        # Prep input data

        pf_w_path = self._get_pinfile_path()
        pf_data_path = self._get_data_path()
        if not pf_data_path:
            pf = self.parser.process(pf_w_path, data=self.pf_data)
        else:
            pf = self.parser.process(pf_w_path,
                                     data='@{0}'.format(pf_data_path))

        if pf:
            provision_data = self._build(pf, pf_data=self.pf_data)

            pf_outfile = self.get_cfg('tmp', 'outfile')
            if pf_outfile:
                self.parser.write_json(provision_data, pf_outfile)

        prov_data = OrderedDict()

        if len(targets):
            for target in targets:
                prov_data[target] = provision_data.get(target)
        else:
            prov_data = provision_data

        return self.do_validation(prov_data, old_schema=old_schema)


    def lp_setup(self, providers=("all")):
        """
        This function takes a list of providers, and setsup the dependencies
        :param providers:
            A tuple of providers to install dependencies
        """
        if self.ctx.get_evar("ask_sudo_pass"):
            output = self._invoke_playbooks(providers=providers,
                                            action="ask_sudo_setup")
        else:
            output = self._invoke_playbooks(providers=providers,
                                            action="setup")

        return output




    def lp_destroy(self, targets=(), run_id=None, tx_id=None):

        """
        This function takes a list of targets, and performs a destructive
        teardown, including undefining nodes, according to the target(s).

        .. seealso:: lp_down - currently unimplemented

        :param targets:
            A tuple of targets to destroy.

        :param run_id:
            An optional run_id to use

        :param tx_id:
            An optional tx_id to use
        """

        # prep inputs


        outputs = self._execute_action('destroy',
                                       targets,
                                       run_id=run_id,
                                       tx_id=tx_id)
        if (('post' in self.pb_hooks) and
            (self.__meta__ == "CLI") and not self.get_cfg('hook_flags',
                                                          'no_hooks')):
            self.hook_state = '{0}{1}'.format('post', 'destroy')
        return outputs


    def _execute_action(self, action, targets=(), run_id=None, tx_id=None):
        """
        This function takes a list of targets, and performs a destructive
        teardown, including undefining nodes, according to the target(s).

        .. seealso:: lp_down - currently unimplemented

        :param targets:
            A tuple of targets to perform action upon.

        :param run_id:
            An optional run_id to use

        :param tx_id:
            An optional tx_id to use
        """

        use_pinfile = True
        pf = None

        return_data = OrderedDict()
        return_code = 0

        urfa = self.get_cfg('lp', 'use_rundb_for_actions')
        use_rundb_for_actions = ast.literal_eval(urfa.title())

        # The UI should catch this, but just in case.
        if run_id and tx_id:
            raise ActionError("'run_id' and 'tx_id' are mutually exclusive")

        if use_rundb_for_actions:
            if run_id and not tx_id:
                if not len(targets) == 1:
                    raise ActionError("'use_rundb_for_actions' is enabled."
                                      " A single target required when"
                                      " passing --run_id.")
                run_id = int(run_id)
                use_pinfile = False
            elif not run_id and tx_id:
                use_pinfile = False

        if use_pinfile:
            pf_w_path = self._get_pinfile_path()
            pf_data_path = self._get_data_path()
            if not pf_data_path:
                pf = self.parser.process(pf_w_path, data=self.pf_data)
            else:
                pf = self.parser.process(pf_w_path,
                                         data='@{0}'.format(pf_data_path))

            if pf:
                provision_data = self._build(pf, pf_data=self.pf_data)

                pf_outfile = self.get_cfg('tmp', 'outfile')
                if pf_outfile:
                    self.parser.write_json(provision_data, pf_outfile)


            return_code, return_data = self._execute(provision_data,
                                                     targets,
                                                     action=action,
                                                     run_id=run_id)
        else:
            # get the pinfile data from the run_id or the tx_id
            provision_data = self.get_pf_data_from_rundb(targets,
                                                         run_id=run_id,
                                                         tx_id=tx_id)

            if provision_data:
                return_code, return_data = self._execute(provision_data,
                                                         targets,
                                                         action=action,
                                                         run_id=run_id,
                                                         tx_id=tx_id)

            else:
                return (99, {})

        return (return_code, return_data)


    def find_include(self, filename, ftype='topology'):
        """
        Find the included file to be acted upon.

        :param filename:
            name of file from to be loaded

        :param ftype:
            the file type to locate: topology, layout
            (default: topology)

        """
        folder = self.get_evar('topologies_folder', 'topologies')
        if ftype == 'layout':
            folder = self.get_evar('layouts_folder', 'layouts')

        path = os.path.realpath('{0}/{1}'.format(self.workspace, folder))
        files = os.listdir(path)

        if filename in files:
            return os.path.realpath('{0}/{1}'.format(path, filename))

        raise LinchpinError('{0} not found in'
                            ' workspace'.format(filename))


    def _make_layout_integers(self, data):

        inv_layout = data.get('inventory_layout')

        if inv_layout:
            hosts = inv_layout.get('hosts')
            for k in hosts:
                count = int(hosts[k].get("count"))
                hosts[k]["count"] = count
            inv_layout["hosts"] = hosts
            data["inventory_layout"] = inv_layout

        return data


    def _build(self, pf, pf_data=None):
        """
        This function constructs the provision_data from the pinfile inputs

        :param pf:
            Provided PinFile dict, with all targets

        """

        provision_data = {}

        for target in pf.keys():
            if target == 'cfgs':
                provision_data['cfgs'] = pf['cfgs']
                continue

            provision_data[target] = {}

            if isinstance(pf[target]['topology'], str):
                topology_path = self.find_include(pf[target]["topology"])
                topology_data = self.parser.process(topology_path,
                                                    data=self.pf_data)
            else:
                topology_data = pf[target]['topology']

            provision_data[target]['topology'] = topology_data

            layout_data = None

            if 'layout' in pf[target]:
                if not isinstance(pf[target]['layout'], dict):
                    layout_path = self.find_include(pf[target]["layout"],
                                                    ftype='layout')

                    layout_data = self.parser.process(layout_path,
                                                      data=self.pf_data)
                    layout_data = self._make_layout_integers(layout_data)
                    provision_data[target]['layout'] = layout_data
                else:
                    layout_data = pf[target]['layout']
                    provision_data[target]['layout'] = layout_data

            if 'hooks' in pf[target]:
                provision_data[target]['hooks'] = pf[target]['hooks']
            # grab target specific vars

        return provision_data


    def _execute(self, provision_data, targets,
                 action='up', run_id=None, tx_id=None):
        """
        This function takes a list of targets, constructs a dictionary of tasks
        and passes it to the LinchpinAPI.do_action method for processing.

        :param provision_data:
            Provided PinFile json data, with available targets

        :param targets:
            A tuple of targets to provision

        :param action:
            Specific action to perform on the provided target(s)

        :param run_id:
            An optional run_id if the task is idempotent or a destroy action
        """

        prov_data = OrderedDict()

        if len(targets):
            for target in targets:
                prov_data[target] = provision_data.get(target)
        else:
            prov_data = provision_data

        return self.do_action(prov_data,
                              action=action,
                              run_id=run_id,
                              tx_id=tx_id)


    def lp_fetch(self, src, root='', fetch_type='workspace',
                 fetch_protocol='FetchGit', fetch_ref=None, dest_ws=None,
                 nocache=False):
        """
        Fetch a workspace from git, http(s), or a local directory, and
        generate a provided workspace

        :param src:         The URL or URI of the remote directory

        :param root:        Used to specify the location of the workspace
                            within the remote. If root is not set, the root
                            of the given remote will be used.

        :param fetch_type:  Specifies which component(s) of a workspace the
                            user wants to fetch. Types include: topology,
                            layout, resources, hooks, workspace.
                            (default: workspace)

        :param fetch_protocol:  The protocol to use to fetch the workspace.
                                (default: git)

        :param fetch_ref:   Specify the git branch. Used only with git protocol
                            (eg. master). If not used, the default branch will
                            be used.

        :param dest_ws: Workspaces destination, the workspace will be relative
                        to this location.

                        If `dest_ws` is not provided and `-r/--root` is
                        provided, the basename will be the name of the
                        workspace within the destination. If no root is
                        provided, a random workspace name will be generated.
                        The destination can also be explicitly set by using
                        -w (see linchpin --help).

        :param nocache: If true, don't copy from the cache dir, unless it's
                        longer than the configured fetch.cache_days (1 day)
                        (default: False)
        """

        root_ws = ''
        if dest_ws and dest_ws != '.':
            if root:
                abs_root = os.path.expanduser(os.path.realpath(root))
                root_ws = os.path.basename(abs_root.rstrip(os.path.sep))
            else:
                # generate a unique value for the root
                hash_string = '{0}{1}'.format(src, dest_ws)
                uroot = hashlib.sha256(hash_string)
                uroot = uroot.hexdigest()[:8]

                # generate a random location to put an underscore
                min_under = randint(1, 7)
                max_under = min_under + 1
                root_ws = uroot[:min_under] + '_' + uroot[max_under:]
        else:
            dest_ws = self.workspace
            if root:
                abs_root = os.path.expanduser(os.path.realpath(root))
                root_ws = os.path.basename(abs_root.rstrip(os.path.sep))
#                else:
#                    pass  # dest = self.workspace (set at the top)

        dest = '{0}/{1}'.format(dest_ws, root_ws)

        output_txt = 'destination workspace: {0}'.format(dest)
        if not os.path.exists(dest):
            os.makedirs(dest)
            output_txt = 'Created {0}'.format(output_txt)

        self.ctx.log_state(output_txt)

        fetch_aliases = {
            "topologies": self.get_evar("topologies_folder"),
            "layouts": self.get_evar("layouts_folder"),
            "resources": self.get_evar("resources_folder"),
            "hooks": self.get_evar("hooks_folder"),
            "workspace": "workspace"
        }

        fetch_type = fetch_aliases.get(fetch_type, 'workspaces')

        cache_path = os.path.abspath(os.path.join(os.path.expanduser('~'),
                                                  '.cache/linchpin'))
        if not os.path.exists(cache_path):
            os.makedirs(cache_path)

        fetch_class = FETCH_CLASS[fetch_protocol](self.ctx, fetch_type, src,
                                                  dest, cache_path, root=root,
                                                  root_ws=root_ws,
                                                  ref=fetch_ref)
        fetch_class.fetch_files()

        if nocache:
            self.set_cfg('fetch', 'cache_ws', 'False')

        fetch_class.copy_files()
